import sys
import types
import contextlib
from functools import wraps
from unittest.case import (
    addModuleCleanup,
    FunctionTestCase,
    SkipTest,
    skip,
    skipIf as skipIfSrc,
    skipUnless,
    expectedFailure,
)
from unittest.case import TestCase as TestCaseSrc
from unittest.case import _Outcome, _ShouldStop, warnings, safe_repr, _OrderedChainMap, _subtest_msg_sentinel

# unittest跟随python版本更新, 新版本unittest存在新的实现, 需要定义最新兼容python版本来防止出现兼容性问题, 并给用户报出来
NEWEST_PYTHON_VERSION = (3, 10, 5)
NEWEST_PYTHON_VERSION_STR = ".".join([str(s) for s in NEWEST_PYTHON_VERSION])

class Outcome(_Outcome):
    def __init__(self, result=None):
        super().__init__(result)
        # 兼容3.11
        self.skipped = []
        self.errors = []
        self.current_stage = None
        self.error_stages = []  # 失败发生在的步骤：_miniClassSetUp, _miniSetUp, _miniTearDown, setUpClass, tearDownClass, setUp, tearDown, testMethod
        self.stage_errors = {}  # 发生在每个stage的具体error

    @contextlib.contextmanager
    def testPartExecutor(self, test_case, isTest=False, error_stage=None):
        """
        rewrite for error_stages
        :error_stage: current error stage name
        """
        old_success = self.success
        self.success = True
        self.current_stage = error_stage or self.current_stage
        try:
            yield
        except KeyboardInterrupt:
            raise
        except SkipTest as e:
            self.success = False
            self.skipped.append((test_case, str(e)))
        except _ShouldStop:
            pass
        except:
            exc_info = sys.exc_info()
            if self.expecting_failure:
                self.expectedFailure = exc_info
            else:
                self.success = False
                self.errors.append((test_case, exc_info))
                if self.current_stage and self.current_stage in self.stage_errors:
                    self.stage_errors[self.current_stage].append(exc_info)
                elif self.current_stage:
                    self.stage_errors[self.current_stage] = [exc_info,]
                    self.error_stages.append(self.current_stage)
            # explicitly break a reference cycle:
            # exc_info -> frame -> exc_info
            exc_info = None
        else:
            if self.result_supports_subtests and self.success:
                self.errors.append((test_case, None))
        finally:
            self.success = self.success and old_success


# 重写testcase
class TestCase(TestCaseSrc):
    def __getattr__(self, name: str):
        if name.startswith("_") and sys.version_info > NEWEST_PYTHON_VERSION:  # 重写框架使用了很多原有的_开头的函数, 如果不存在, 有可能是因为python版本引起的
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}' maybe because your python version is greater than {NEWEST_PYTHON_VERSION_STR}")
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def _miniSetUp(self):
        pass

    def _miniTearDown(self):
        pass

    def run(self, result=None):
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, "startTestRun", None)
            if startTestRun is not None:
                startTestRun()

        result.startTest(self)

        testMethod = getattr(self, self._testMethodName)
        if getattr(self.__class__, "__unittest_skip__", False) or getattr(
            testMethod, "__unittest_skip__", False
        ):
            # If the class or method was skipped.
            try:
                skip_why = getattr(
                    self.__class__, "__unittest_skip_why__", ""
                ) or getattr(testMethod, "__unittest_skip_why__", "")
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return

        expecting_failure_method = getattr(
            testMethod, "__unittest_expecting_failure__", False
        )
        expecting_failure_class = getattr(self, "__unittest_expecting_failure__", False)
        expecting_failure = expecting_failure_class or expecting_failure_method
        outcome = Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self, error_stage="_miniSetUp"):
                self._miniSetUp()
            if outcome.success:
                with outcome.testPartExecutor(self, error_stage="setUp"):
                    self._callSetUp()
                if outcome.success:
                    outcome.expecting_failure = expecting_failure
                    with outcome.testPartExecutor(
                        self, isTest=True, error_stage="testMethod"
                    ):
                        self._callTestMethod(testMethod)
                    outcome.expecting_failure = False
                    with outcome.testPartExecutor(self, error_stage="tearDown"):
                        self._callTearDown()
            outcome.expecting_failure = False
            with outcome.testPartExecutor(self, error_stage="_miniTearDown"):
                self._miniTearDown()

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, "stopTestRun", None)
                if stopTestRun is not None:
                    stopTestRun()

            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome -> outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None

            # clear the outcome, no more needed
            self._outcome = None

    @contextlib.contextmanager
    def subTest(self, msg=_subtest_msg_sentinel, **params):
        """Return a context manager that will return the enclosed block
        of code in a subtest identified by the optional message and
        keyword parameters.  A failure in the subtest marks the test
        case as failed but resumes execution at the end of the enclosed
        block, allowing further test code to be executed.
        """
        if self._outcome is None or not self._outcome.result_supports_subtests:
            yield
            return
        parent = self._subtest
        if parent is None:
            params_map = _OrderedChainMap(params)
        else:
            params_map = parent.params.new_child(params)
        self._subtest = _SubTest(self, msg, params_map)
        try:
            with self._outcome.testPartExecutor(self._subtest, isTest=True):
                yield
            if not self._outcome.success:
                result = self._outcome.result
                if result is not None and result.failfast:
                    raise _ShouldStop
            elif self._outcome.expectedFailure:
                # If the test is expecting a failure, we really want to
                # stop now and register the expected failure.
                raise _ShouldStop
        finally:
            self._subtest = parent

    def assertSetContainsSubset(self, subset: set, superset: set, msg=None):
        """Checks whether superset is a superset of subset."""
        warnings.warn("assertSetContainsSubset is deprecated", DeprecationWarning)
        missing = []
        for item in subset:
            if item not in superset:
                missing.append(item)

        if not missing:
            return

        standardMsg = ""
        if missing:
            standardMsg = "Missing: %s" % ",".join(safe_repr(m) for m in missing)

        self.fail(self._formatMessage(msg, standardMsg))


class _SubTest(TestCase):

    def __init__(self, test_case, message, params):
        super().__init__()
        self._message = message
        self.test_case = test_case
        self.params = params
        self.failureException = test_case.failureException

    def runTest(self):
        raise NotImplementedError("subtests cannot be run directly")

    def _subDescription(self):
        parts = []
        if self._message is not _subtest_msg_sentinel:
            parts.append("[{}]".format(self._message))
        if self.params:
            params_desc = ', '.join(
                "{}={!r}".format(k, v)
                for (k, v) in self.params.items())
            parts.append("({})".format(params_desc))
        return " ".join(parts) or '(<subtest>)'

    def id(self):
        return "{} {}".format(self.test_case.id(), self._subDescription())

    def shortDescription(self):
        """Returns a one-line description of the subtest, or None if no
        description has been provided.
        """
        return self.test_case.shortDescription()

    def __str__(self):
        return "{} {}".format(self.test_case, self._subDescription())


def _whether_need_skip(func: types.FunctionType, *args, **kwargs):
    if func.__code__.co_argcount == 0:
        need_skip = func()
    elif func.__code__.co_argcount == 1:
        need_skip = func(args[0])  # just use TestCase
    else:
        need_skip = func(*args, **kwargs)  # mybe ddt case
    return need_skip

def skipIf(func_or_condition: types.FunctionType or bool, reason=''):
    """
    Skip a test if the condition is true.
    """
    if isinstance(func_or_condition, bool):
        return skipIfSrc(func_or_condition, reason)
    def wrapper(wrapped):
        @wraps(wrapped)
        def inner(*args, **kwargs):
            if len(args) == 0:  # not TestCase
                return wrapped(*args, **kwargs)
            if isinstance(args[0], TestCaseSrc):  # is TestCase
                self = args[0]
                if _whether_need_skip(func_or_condition, *args, **kwargs):
                    self.skipTest(reason)
                return wrapped(*args, **kwargs)
            # normal wrapped func
            if _whether_need_skip(func_or_condition, *args, **kwargs):
                raise SkipTest(reason)
            return wrapped(*args, **kwargs)
        return inner
    return wrapper

def expectedException(exception_or_type: Exception or type):
    def wrapper(wrapped):
        @wraps(wrapped)
        def inner(self: TestCase, *args, **kwargs):
            if isinstance(exception_or_type, type):
                _type = exception_or_type
                _value = None
            else:
                _type = exception_or_type.__class__
                _value = exception_or_type.__repr__()
            with self.assertRaises(_type) as cm:
                wrapped(self, *args, **kwargs)
            if _value is not None:
                self.assertEqual(_value, cm.exception.__repr__())
            # try:
            #     return wrapped(self, *args, **kwargs)
            # except _type as t:
            #     if _value is None or _value == t.__repr__():
            #         return
            #     raise t
            # else:
            #     raise
        return inner
    return wrapper


'''
Author: yopofeng yopofeng@tencent.com
Date: 2023-04-18 15:36:11
LastEditors: yopofeng yopofeng@tencent.com
LastEditTime: 2023-04-25 15:18:04
FilePath: /py-minium/minium/native/wx_native/wording.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''
from __future__ import annotations
from enum import Enum
import typing

class Language(Enum):
    zh = 1
    en = 2

class Item(object):
    def __init__(self, value: typing.Dict[Language, typing.Any]) -> None:
        self._value = {
            k.name: v for k, v in value.items()
        }

    @property
    def value(self):
        return self._value[WORDING.language.name]
    
    def __getattr__(self, name):
        if name in self._value:
            return self._value[name]
        raise

class Meta(type):
    def __new__(cls, name, base, attr_dict: typing.Dict[str, typing.Dict[Language, typing.Any]]):
        for k, v in attr_dict.items():
            if k.startswith("__"):
                continue
            attr_dict[k] = Item(v)
        return super().__new__(cls, name, base, attr_dict)
        

class WORDING:
    language = Language.zh

    @classmethod
    def setLanguage(cls, value: Language):
        cls.language = value

    class COMMON(object, metaclass=Meta):
        LOGIN_WORDS = {
            Language.zh: {"微信", "通讯录", "发现", "我"},
            Language.en: {"Chats", "Contacts", "Discover", "Me"}
        }

        MODAL_CONFIRM = {
            Language.zh: "确定",
            Language.en: "OK"
        }

        MODAL_CANCEL = {
            Language.zh: "取消",
            Language.en: "CANCEL"
        }
        
        LOCAL_DEBUG_MODAL_TITLE = {
            Language.zh: "本地调试已结束",
            Language.en: "Local debugging ended"
        }

        LOCAL_DEBUG_MODAL_TITLE2 = {
            Language.zh: "确认结束调试?",
            Language.en: "确认结束调试?"
        }
        
        MORE: Item = {
            Language.zh: "更多",
            Language.en: "More"
        }
        
        SEARCH: Item = {
            Language.zh: "搜索",
            Language.en: "Search"
        }

        DONE: Item = {
            Language.zh: "完成",
            Language.en: "Done"
        }

        SEND: Item = {
            Language.zh: "发送",
            Language.en: "Send"
        }



    class IOS(object, metaclass=Meta):
        # 转发/分享
        FORWARD_WORD1: Item = {
            Language.zh: "转发",
            Language.en: "Forward"
        }
        FORWARD_WORD2: Item = {
            Language.zh: "转发给朋友",
            Language.en: "Send to Chat"
        }
        FORWARD_WORD3: Item = {
            Language.zh: "发送",
            Language.en: "Send"
        }
        CREATE_CHAT1: Item = {
            Language.zh: "创建聊天",
            Language.en: "New Chat"
        }
        CREATE_CHAT2: Item = {
            Language.zh: "创建新的聊天",
            Language.en: "New Chat"
        }
        


    class ANDROID(object, metaclass=Meta):
        pass

if __name__ == "__main__":
    print(WORDING.COMMON.LOGIN_WORDS.zh)



# -*- coding: utf-8 -*-
"""
@author: 'xiazeng'
@created: 2017/3/31 
"""
import os.path

AT_ROOT = os.path.dirname(os.path.dirname(__file__))
JAR_STUB_FILENAME = "AtStub.jar"
JAR_STUB_PATH = os.path.join(AT_ROOT, "bin", JAR_STUB_FILENAME)
JAR_STUB_CLASS = "com.tencent.mm.test.at.Stub"
STUB_CASE_NAME = "testStartServer"

STUB_APK_PATH = os.path.join(AT_ROOT, "bin", "AtServer.apk")
TEST_APK_ACT = "ui.LauncherUI"
TEST_APK_SERVICE = "ui.ApiService"

TEST_APP_CLS = "com.tencent.weautomator.at.Stub"
TEST_APP_PKG = "com.tencent.weautomator"

STF_LIB_PATH = os.path.join(AT_ROOT, "bin", "stf_libs")

# version
UIAUTOMATOR2 = 2



"""
callback类, 供hook等需要callback的场景使用
"""
import types
import threading
import sys

from ...framework.exception import MiniTimeoutError
from ...utils.eventloop import event_loop
from ...utils.utils import async_wait, WaitTimeoutError, Future, EventLoop


class Callback(object):
    def __init__(self, callback: types.FunctionType = None) -> None:
        self.__callback = callback
        self.__called = threading.Semaphore(0)
        self.__is_called = False
        self.__callback_result = None

    def callback(self, args):
        self.__is_called = True
        self.__callback_result = (
            args[0] if isinstance(args, (tuple, list)) and len(args) == 1 else args
        )
        self.__called.release()
        if self.__callback:
            self.__callback(*args)

    @property
    def is_called(self):
        """callback曾被调用过

        :return bool: True: called, False: never called
        """
        return self.__is_called

    @property
    def result(self):
        return self.__callback_result

    def acquire(self, timeout=10):
        """acquire next callback

        :param int timeout: wait seconds, defaults to 10
        """
        return self.__called.acquire(timeout=timeout)

    def wait_called(self, timeout=10) -> bool:
        """
        等待回调调用, 默认等待最多10s
        """
        if self.__is_called:
            return True
        return self.acquire(timeout=timeout)

    def get_callback_result(self, timeout=0) -> any:
        """
        获取回调结果, 超时未获取到结果报AssertionError
        1. 回调参数只有一个的情况会解构
        2. 回调参数中有多个的情况会直接返回参数list
        """
        if self.wait_called(timeout):
            return self.__callback_result
        assert self.__is_called, f"No callback received within {timeout} seconds"


class AsyncCallback(object):
    def __init__(self, loop: EventLoop = None) -> None:
        if loop is None:
            self._loop = event_loop
        else:
            self._loop = loop
        self._is_called = False
        self._waiter: Future = self._loop.create_future()

    async def set_result(self, args):
        if not self._waiter.done():
            self._is_called = True
            result = args[0] if args and len(args) == 1 else args
            if isinstance(result, BaseException):
                self._waiter.set_exception(result)
            else:
                self._waiter.set_result(result)

    def cancel(self):
        self._waiter.cancel()

    def callback(self, *args):
        if not self._waiter.done():
            self._loop.run_coroutine(self.set_result(args))

    @property
    def is_called(self):
        return self._waiter.done()

    def wait_called(self, timeout=10) -> bool:
        """
        等待回调调用, 默认等待最多10s
        """
        if timeout == 0:
            return self._waiter.done()
        if self._waiter.done():
            return True
        try:
            return self._loop.run_coroutine(
                async_wait(self._waiter, timeout, self._loop)
            ).result()
        except WaitTimeoutError:
            return False
        except:
            return self._waiter.done()

    def get_callback_result(self, timeout=0) -> any:
        if self.wait_called(timeout):
            try:
                return self._waiter.result()
            except:
                return sys.exc_info()[1]
        assert self._is_called, f"No callback received within {timeout} seconds"

    def get_result(self, timeout=0):
        """
        获取回调结果, 结果如果为exception直接抛出, 超时未获取到则抛MiniTimeoutError
        """
        if self.wait_called(timeout):
            return self._waiter.result()
        raise MiniTimeoutError(f"No callback received within {timeout} seconds")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2018/3/14
import threading
import logging
import time


logger = logging.getLogger()


class ThreadHelper(threading.Thread):
    def __init__(self, *args, **kwargs):
        super(ThreadHelper, self).__init__(*args, **kwargs)
        self.thread_main = None
        self.args = None
        self.kwargs = None
        self.thread_result = None
        self.run_exception = None

    @property
    def result(self):
        self.join()
        if self.run_exception is not None:
            raise self.run_exception
        return self.thread_result

    def run(self):
        try:
            s = time.time()
            self.thread_result = self.thread_main(*self.args, **self.kwargs)
            logger.info("costs %.2fms", time.time()-s)
        except Exception as e:
            logger.exception("failed")
            self.run_exception = e


def run(f, *args, **kwargs):
    thread_helper = ThreadHelper()
    thread_helper.args = args
    thread_helper.kwargs = kwargs
    thread_helper.thread_main = f
    thread_helper.setDaemon(True)
    thread_helper.start()
    return thread_helper


def post_run(delay, f, *args, **kwargs):
    time.sleep(delay)
    run(f, *args, **kwargs)


def test(a):
    import time
    time.sleep(1)
    return a

if __name__ == '__main__':
    a = run(test, 1)
    b = run(test, 'ad')

    print(b.result)
    print(a.result)



from unittest.suite import TestSuite as TestSuiteSrc
from unittest.suite import BaseTestSuite
from unittest.suite import _call_if_exists, _DebugResult
import sys
from . import case


# 重写TestSuite
class TestSuite(TestSuiteSrc):
    def __str__(self):
        s = ""
        for test in self._tests:
            s += "%s\n" % str(test)
        return s

    def _miniClassSetUp(self, test, result):
        pass

    def _createClassFailure(self, test, result, exception=None, info=None, error_stages=None):
        addSkip = getattr(result, "addSkip", None)
        if addSkip is not None and isinstance(exception, case.SkipTest):
            addSkip(test, str(exception))
        else:
            result.startTest(test)
            addErrorStage = getattr(result, "addErrorStage", None)
            if addErrorStage is not None:
                result.addErrorStage(test, info or sys.exc_info(), error_stages)
            else:
                result.addError(test, info or sys.exc_info())
            result.stopTest(test)

    def _handleClassSetUp(self, test, result):
        previousClass = getattr(result, "_previousTestClass", None)
        currentClass = test.__class__
        if currentClass == previousClass:
            if getattr(currentClass, "__class_setup_failure_why__", None):
                self._createClassFailure(
                    test,
                    result,
                    info=getattr(currentClass, "__class_setup_failure_why__", None),
                    error_stages=getattr(currentClass, "__class_setup_failure_stages__")
                )
            return
        if result._moduleSetUpFailed:
            return
        if getattr(currentClass, "__unittest_skip__", False):
            return

        try:
            currentClass._classSetupFailed = False
        except TypeError:
            # test may actually be a function
            # so its class will be a builtin-type
            pass

        error_stages = []
        stage_errors = {}
        # run _miniClassSetUp
        _miniClassSetUp = getattr(currentClass, "_miniClassSetUp", None)
        setattr(currentClass, "__is_minitest_suite__", True)
        if _miniClassSetUp is not None:
            _call_if_exists(result, "_setupStdout")
            try:
                _miniClassSetUp()
            except Exception as e:
                if isinstance(result, _DebugResult):
                    raise
                currentClass._classSetupFailed = True
                error_stages.append("_miniClassSetUp")
                info = sys.exc_info()
                stage_errors["_miniClassSetUp"] = [info,]
                setattr(currentClass, "__class_setup_failure_why__", info)
                setattr(currentClass, "__class_setup_failure_stages__", error_stages)
                setattr(currentClass, "__class_setup_failure_errors__", stage_errors)
                className = "%s.%s" % (
                    currentClass.__module__,
                    currentClass.__qualname__,
                )
                self._createClassFailure(test, result, e, info, error_stages)
                # self._createClassOrModuleLevelException(result, e,
                #                                         '_miniClassSetUp',
                #                                         className)
                info = None
                return
            finally:
                _call_if_exists(result, "_restoreStdout")
                if currentClass._classSetupFailed is True:
                    currentClass.doClassCleanups()
                    if len(currentClass.tearDown_exceptions) > 0:
                        for exc in currentClass.tearDown_exceptions:
                            self._createClassOrModuleLevelException(
                                result, exc[1], "setUpClass", className, info=exc
                            )
                    return

        # run setUpClass
        setUpClass = getattr(currentClass, "setUpClass", None)
        if setUpClass is not None:
            _call_if_exists(result, "_setupStdout")
            try:
                setUpClass()
            except Exception as e:
                if isinstance(result, _DebugResult):
                    raise
                currentClass._classSetupFailed = True
                error_stages.append("setUpClass")
                info = sys.exc_info()
                stage_errors["_miniClassSetUp"] = [info,]
                setattr(currentClass, "__class_setup_failure_why__", info)
                setattr(currentClass, "__class_setup_failure_stages__", error_stages)
                setattr(currentClass, "__class_setup_failure_errors__", stage_errors)
                className = "%s.%s" % (
                    currentClass.__module__,
                    currentClass.__qualname__,
                )
                self._createClassFailure(test, result, e, info, error_stages)
                # self._createClassOrModuleLevelException(result, e,
                #                                         'setUpClass',
                #                                         className)
                info = None
            finally:
                _call_if_exists(result, "_restoreStdout")
                if currentClass._classSetupFailed is True:
                    currentClass.doClassCleanups()
                    if len(currentClass.tearDown_exceptions) > 0:
                        for exc in currentClass.tearDown_exceptions:
                            self._createClassOrModuleLevelException(
                                result, exc[1], "setUpClass", className, info=exc
                            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Author: yopofeng yopofeng@tencent.com
Date: 2022-10-12 11:50:48
LastEditors: yopofeng yopofeng@tencent.com
LastEditTime: 2023-03-27 11:48:24
FilePath: /py-minium/minium/miniprogram/base_driver/minium.py
Description: minium基类
"""
from .app import App
from .connection import Connection
from .minium_object import MiniumObject
from ...framework.miniconfig import MiniConfig
from ...native import BaseNative


class BaseMinium(MiniumObject):
    """
    自动化入口
    """

    app: App
    connection: Connection
    conf: MiniConfig
    native: BaseNative
    platform: str
    is_app_relaunching: bool = False

    def __init__(
        self,
        conf: MiniConfig = None,
        uri="ws://localhost",
        native: BaseNative = None,
        **kwargs,
    ):
        super().__init__()

    # framework use
    def wait_app_relaunch(self) -> None or Exception:
        ...

    def clear_auth(self):
        """清除用户授权信息"""
        ...

    def get_system_info(self) -> dict:
        """获取系统信息

        :return dict: 同小程序接口[wx.getSystemInfo](https://developers.weixin.qq.com/miniprogram/dev/api/base/system/wx.getSystemInfo.html)
        """
        ...

    def enable_remote_debug(self, use_push=True, path=None, connect_timeout=180):
        """在真机上运行自动化

        :param bool use_push: 是否直接推送到客户端, defaults to True
        :param str path: 远程调试二维码的保存路径, defaults to None
        :param int connect_timeout: 连接超时时间，单位 s, defaults to 180
        """
        ...

    def reset_remote_debug(self):
        """重置远程调试，解决真机调试二维码扫码界面报-50003 等常见错误。一般供测试框架调用"""
        ...

    def shutdown(self):
        """测试结束时调用, 停止 微信开发者IDE 以及 minium, 并回收资源。一般供测试框架调用"""
        ...



#!/usr/bin/env python3
# Created by xiazeng on 2019/3/12
import base64
import socket
import sys


def base64_to_file(base64_str, path):
    image_data = base64.b64decode(base64_str)
    f = open(path, "wb")
    f.write(image_data)
    f.close()



def pick_unuse_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    port = s.getsockname()[1]
    s.close()
    return port


def get_std_encoding(stream):
    return getattr(stream, "encoding", None) or sys.getfilesystemencoding()



# -*- coding: utf-8 -*-
"""
@author: 'xiazeng'
@created: 2017/5/3 
"""
import json

from .core import element
from .core import javadriver

try:
    from urllib import unquote
except ImportError:
    from urllib.parse import unquote

g_event_index = 0


class EventMonitor(object):
    def __init__(self, jd):
        self.jd = jd  # type: javadriver.JavaDriver

    def add_event_click(self, match_element, action_element):
        self.add_selector_filter(None, match_element, action_element)

    def add_selector_filter(self, name, match_element, action_element):
        """
        用于处理弹框提示，在match_element出现的时候对action_element进行点击，name是个可以
        :type name: str
        :type match_element: element.Element
        :type action_element: element.Element
        """
        if name is None:
            global g_event_index
            g_event_index += 1
            name = f"EventMonitor_{g_event_index}"
        params = {
            "MatchSelector": unquote(json.dumps(match_element._selector)),
            "ActionSelector": unquote(json.dumps(action_element._selector)),
            "name": name
        }
        self.jd.add_request_when_reconnect('AddEventMonitor', http_params=params)
        self.jd.do_request("AddEventMonitor", params)

    def sync_event(self):
        self.jd.do_request("syncEventMonitor", {})

    def clear(self):
        self.jd.do_request("clearEventMonitor", {})

    def action_when_screen(self, true_or_false):
        self.jd.do_request("setEventMonitorShouldScreen", {"should": true_or_false})

    def get_all_screen(self):
        return self.jd.do_request("getMonitorScreen", {})

    def clear_all_screen(self):
        return self.jd.do_request("clearMonitorScreen", {})

    def remove_selector_filter(self, name):
        params = {
            "name": name
        }
        self.jd.do_request("removeEventMonitor", params)

    def importPath(self, str):
        params = {
            "name": str
        }
        self.jd.do_request("importPath", params)

    def on_off(self, on):
        params = {
            "name": on
        }
        self.jd.do_request("on_off", params)


#!/usr/bin/env python3
# Created by xiazeng on 2019-05-22
import os.path
import time

from minium.utils.utils import timeout, wait
from .basenative import BaseNative, NativeError, ResetError, ModalStatus
import at
from at.core.adbwrap import AdbWrap
from at.core.element import Element as ATElement, AtSelector
import at.keycode
import json
import logging
import threading
import shutil
import ssl
import requests
import re
from enum import Enum
from ..lib.android_base import UiDefine, ScreenType
from .wording import Language, WORDING

ssl._create_default_https_context = ssl._create_unverified_context

WECHAT_PACKAGE = "com.tencent.mm"
WECHAT_ACTIVITY = "ui.LauncherUI"
MINIPROGRAM_ACTIVITY = "plugin.appbrand.ui.AppBrandUI"
MINIPROGRAM_PLUGIN_ACTIVITY = "plugin.appbrand.ui.AppBrandPluginUI"


logger = logging.getLogger("minium")
_MIN_NORMALIZED_FRAME_LENGTH = 0.5


class WXAndroidNative(BaseNative):
    def __init__(self, json_conf):
        super(WXAndroidNative, self).__init__(json_conf)
        if json_conf is None:
            json_conf = {}
        self.serial = json_conf.get("serial") or AdbWrap.get_default_serial()
        uiautomator_version = int(json_conf.get("uiautomator_version", "2"))
        at.uiautomator_version = uiautomator_version
        self.at = at.At(self.serial)
        self.at.java_driver.set_capture_op(False)
        self.ui = UiDefine(self.at)
        self.lang = json_conf.get("lang")
        self.perf_thread = None
        self.perf_flag = False
        self.ef_able_flag = False
        self.perf_data = []
        self.fps_data_dict = {}  # fps dict
        self.jank_count_dict = {}  # 卡顿次数
        self.fps_thread = None
        self.startup_time = 0
        self.outputs = json_conf.get("outputs") or os.path.dirname(
            os.path.realpath(__file__)
        )
        self.debug = json_conf.get("debug", False)
        self.outputs_screen = os.path.join(self.outputs, "image")
        self.screen_type = ScreenType.AT  # 截图方式，默认使用at
        self._empty_base_screen_dir(self.outputs_screen)
        self._current_views = None
        # 1. 原始版本, 使用计算数字坐标的方式进行点击. 
        # 2. 数字按钮具有固定的rid, 格式为: `com.tencent.mm:id/tenpay_keyboard_{number}`
        self.pay_modal_version = 0

    @property
    def current_views(self):
        if self._current_views is None:
            return self._get_current_views()
        return self._current_views

    @timeout(2, 0.5)
    def _get_current_views(self):
        self._current_views = self.at.java_driver.dump_all_views()
        return self._current_views

    def _el_exist_in_current_view(self, el: ATElement):
        if not self.current_views:
            return False
        instance = el._selector.get(AtSelector.SELECTOR_INSTANCE, 0)
        for ui_view_list in self.current_views:
            for ui_view in ui_view_list:
                if el.match_ui_view(ui_view):
                    if instance == 0:
                        return True
                    else:
                        instance -= 1
        return False

    def _empty_base_screen_dir(self, dirname):
        if os.path.exists(dirname):
            if os.path.isdir(dirname):
                shutil.rmtree(dirname)
                time.sleep(1)
            else:
                os.remove(dirname)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

    def _deal_log(self, log):
        log_2 = ""
        log_4 = ""
        for line in log:
            if str(line).find(",StartUp,0") != -1:
                log_1 = str(line).split(",StartUp,0")[0]
                log_2 = str(log_1).split(",")[-1]
                logger.info(log_2)
            elif str(line).find(",StartUp,1") != -1:
                log_3 = str(line).split(",StartUp,1")[0]
                log_4 = str(log_3).split(",")[-1]
                logger.info(log_4)
            else:
                pass

        if log_2 and log_4:
            startup_time = int(log_4) - int(log_2)
            self.startup_time = startup_time
            return startup_time
        else:
            return 0

    @property
    def logcat(self):
        return self.at.logcat

    def connect_weapp(self, path):
        self.at.apkapi.launch()
        gallery_name = "atstub"
        self.at.apkapi.add_gallery(path)
        self.stop_wechat()
        self.start_wechat()
        if self.lang == "en":
            self.e.text("Discover").click()
            self.e.text("Scan").click()
            # self.e.cls_name("android.widget.ImageButton").click()
            # self.e.text("Choose QR Code from Album").click()
            self.e.cls_name("android.widget.LinearLayout").instance(1).click()
            self.e.text("All Images").click()
            self.e.text(gallery_name).click()
            self.e.desc_contains("Image 1").click()
            self.e.text_contains("Scanning").wait_disappear()
        else:
            self.e.text("发现").click()
            self.e.text("扫一扫").click()
            self.e.cls_name("android.widget.LinearLayout").instance(1).click()
            # self.e.cls_name("android.widget.ImageButton").click()
            # self.e.text("从相册选取二维码").click()
            self.e.text("所有图片").click()
            self.e.text(gallery_name).click()
            self.e.desc_contains("图片 1").click()
            self.e.text_contains("扫描中").wait_disappear()

    def screen_shot(self, filename, quality=30):
        return self._screen_shot(filename, quality)

    def _screen_shot(self, filename, quality=30):
        """
        安卓截图有两种方式：
        1. via at
        2. via adb screencap
        via at现存问题: 可能因为网络、at进程被kill等问题, 截图时间超长
        """
        if self.screen_type == ScreenType.AT:
            stime = time.time()
            try:
                ret = self.at.device.screen_shot(filename, quality=quality)
            except (requests.ConnectionError, requests.ReadTimeout) as e:
                logger.error("screen_shot error %s, try reconnect" % str(e))
                self.at.java_driver.reconnect()
                stime = time.time()
                ret = self.at.device.screen_shot(filename, quality=quality)
            # ret = self.at.device.screen_shot(filename, quality=quality)
            if time.time() - stime > 50:  # 如果一次截图超过50s, 就切换截图类型
                logger.warning(
                    "screen_shot cost bigger than 50s, change the screen type"
                )
                self.screen_type = ScreenType.ADB
            return ret
        elif self.screen_type == ScreenType.ADB:
            return self.at.device.screen_shot(filename, quality=quality, display_id=0)

    def pick_media_file(
        self,
        cap_type="camera",
        media_type="photo",
        original=False,
        duration=5.0,
        index_list=None,
    ):
        pass

    def input_text(self, text):
        """ """
        self.at.e.edit_text().enter(text, is_click=False)

    def input_clear(self):
        """
        input 组件清除文字(使用此函数之前必须确保输入框处于输入状态)
        :return:
        """
        self.at.e.edit_text().enter("", is_click=False)

    def textarea_text(self, text, index=0):
        self.at.e.edit_text().instance(index).enter(text, is_click=False)

    def textarea_clear(self, index=0):
        self.at.e.edit_text().instance(index).enter("", is_click=False)

    def map_select_location(self, name=None):
        if not name:
            # 不需要搜索地址
            if self.e.text("完成").enabled(True).exists(3):
                return self.e.text("完成").enabled(True).click_if_exists()
            else:  # 不能点完成，返回
                self.map_back_to_mp()
                return False
        can_search = True
        if (
            not self.at.e.text("搜索地点").click_if_exists()
            and not self.e.rid(
                "com.tencent.mm:id/send_loc_search_hint"
            ).click_if_exists()
        ):
            logger.error("当前客户端不支持搜索地点")
            can_search = False
        if not can_search:  # 同不需要搜索
            return self.map_select_location(None)
        self.at.e.edit_text().enter_chinese(name)
        self.at.adb.press_search()
        self.at.adb.press_back()
        while (
            self.at.e.cls_name("android.widget.TextView").text_contains(name).exists()
        ):
            try:
                self.at.e.cls_name("android.widget.TextView").text_contains(
                    name
                ).click()
                if self.at.e.text("完成").exists:
                    break
            except Exception as e:
                logger.warning(str(e))
        self.at.e.text("完成").click()

    def map_back_to_mp(self):
        if not self.at.e.rid("com.tencent.mm:id/btn_back_txt").click_if_exists():
            self.e.cls_name("android.widget.TextView").text("取消").click()

    def deactivate(self, duration):
        self.at.adb.press_home()
        time.sleep(duration)
        self.at.adb.start_app("com.tencent.mm", "ui.LauncherUI", "-n")
        time.sleep(2)
        screen_width = int(self.at.device.width())
        screen_height = int(self.at.device.height())
        self.at.device.swipe(
            int(screen_width / 2),
            int(screen_height * 1 / 4),
            int(screen_width / 2),
            int(screen_height * 4 / 5),
            10,
        )
        time.sleep(2)
        self.textbounds = self.at.e.rid("com.tencent.mm:id/test_mask").get_bounds()
        print(self.textbounds)
        xx = self.textbounds[0]
        yy = self.textbounds[1]
        self.at.adb.click_point(xx, yy)

    def allow_authorize(self, answer=True):
        # 点击按钮触发弹窗到客户端弹窗可能有一定延迟
        if answer:
            return self._handle_btn("允许", timeout=5.0)
        else:
            return self._handle_btn("拒绝", timeout=5.0)

    def allow_login(self, answer=True):
        self.allow_authorize(answer)

    def allow_get_user_info(self, answer=True):
        self.allow_authorize(answer)

    def allow_get_location(self, answer=True):
        self.allow_authorize(answer)

    def allow_get_we_run_data(self, answer=True):
        """
        处理获取微信运动数据确认弹框, 测试号可能需要开通微信运动
        :param answer: True or False
        :return:
        """
        if (
            self.e.cls_name("android.widget.TextView")
            .text("开启微信运动")
            .wait_exists(timeout=3.0)
        ):
            if not answer:
                self.at.adb.press_back()
                return False
            self.e.cls_name("android.widget.TextView").text("启用该功能").click()
        self.allow_authorize(answer)
        return True

    def allow_record(self, answer=True):
        self.allow_authorize(answer)

    def allow_write_photos_album(self, answer=True):
        """
        处理保存相册确认弹框
        :param answer: True or False
        :return:
        """
        self.allow_authorize(answer)

    def allow_camera(self, answer=True):
        """
        处理使用摄像头确认弹框
        :param answer: True or False
        :return:
        """
        self.allow_authorize(answer)

    def allow_get_user_phone(self, answer=True):
        """
        处理获取用户手机号码确认弹框
        :param answer: True or False
        :return:
        """
        self.allow_authorize(answer)

    def allow_send_subscribe_message(self, answer=True):
        """
        允许发送订阅消息
        """
        if answer:
            return self._handle_btn("", timeout=5.0, text_matches="允许|确定")
        else:
            return self._handle_btn("取消", timeout=5.0)

    def _handle_btn(self, btn_text="确定", timeout=0.5, text_matches=None):
        logger.debug("handle android.widget.Button[%s]" % (text_matches or btn_text))

        def do():
            e = self.e.cls_name("android.widget.Button").focusable(True)
            if text_matches:
                e.text_matches(text_matches)
            else:
                e.text(btn_text)
            return e.click_if_exists(timeout)

        try:
            ret = do()
        except (requests.ConnectionError, requests.ReadTimeout) as e:
            logger.error("handle btn error %s" % str(e))
            self.at.java_driver.reconnect()
            ret = do()
        except at.core.exceptions.UiNotFoundError:
            return False
        return ret

    def handle_modal(self, btn_text="确定", title=None, index=-1, force_title=False):
        if title:
            if not self.at.e.text_contains(title).exists():
                logger.info("没有出现预期弹窗：title[%s]", title)
                return False
        ret = self._handle_btn(btn_text)
        return ret

    def handle_action_sheet(self, item):
        ret = self.e.cls_name("android.widget.TextView").text(item).click()
        return ret

    def forward_miniprogram(self, name, text=None, create_new_chat=True):
        self.ui.action_menu.click()
        if not self.e.text("发送给朋友").click_if_exists():
            assert self.e.text("转发给朋友").click_if_exists(0.5), "找不到转发按钮"
        return self.forward_miniprogram_inside(name, text, create_new_chat)

    def _create_new_chat(self, timeout=0.5):
        # 点击"创建新聊天"/"创建聊天"/"创建新的聊天"
        return self.e.text_matches("创建(新的)?聊天").click_if_exists(timeout)

    def forward_miniprogram_inside(self, name, text=None, create_new_chat=True):
        if create_new_chat and self._create_new_chat():
            self.e.edit_text().enter(name)
            time.sleep(1)
            self.e.text_contains(name).click(True)
            if self.e.text("确定(1)").exists():
                self.e.text("确定(1)").enabled(True).click()
            else:
                self.e.text("完成(1)").enabled(True).click()
        else:
            self.e.text("搜索").click()
            time.sleep(1)
            self.e.edit_text().enter(name)
            time.sleep(1)
            self.e.text_contains(name).click(True)
            time.sleep(1)
        if text:
            self.e.edit_text().enter(text)
            self.at.adb.press_back()  # 键盘会挡住

        self.e.cls_name("android.widget.Button").text("发送").click()
        self.e.text("已转发").wait_disappear()

    def send_custom_message(self, message=None):
        pass

    def call_phone(self):
        pass

    def handle_picker(self, *items):
        instance = 0
        for item in items:
            input_elem = self.ui.comp_picker_input.instance(instance)
            next_elem = input_elem.parent().child("android.widget.Button")
            first_text = input_elem.get_text()
            while True:
                # todo: 要判断上滑还是下滑
                current_text = input_elem.get_text()
                if current_text == str(item):
                    break
                if first_text == str(item):
                    raise NativeError(" not found")
            instance += 1

    def get_authorize_settings(self):
        """
        在小程序的授权页面，获取小程序的授权设置
        :return:
        """
        ui_views = self.at.java_driver.dump_ui()
        setting_map = {}
        for ui_view in ui_views:
            if ui_view.cls_name == "android.view.View" and ui_view.content_desc in [
                "已开启",
                "已关闭",
            ]:
                check_status = True if ui_view.content_desc == "已开启" else False
                parant_view = ui_view.sibling().get_children()[0]
                setting_map[parant_view.text] = check_status
        return setting_map

    def back_from_authorize_setting(self):
        """
        从小程序授权页面跳转回小程序
        :return:
        """
        self.at.adb.press_back()

    def authorize_page_checkbox_enable(self, name, enable):
        """
        在小程序授权设置页面操作CheckBox
        :param name: 设置的名称
        :param enable: 是否打开
        :return:
        """
        setting_map = self.get_authorize_settings()
        if setting_map.get(name) == enable:
            return
        self.e.text(name).parent().instance(2).child().cls_name(
            "android.view.View"
        ).click()
        if not enable:
            self.e.cls_name("android.widget.Button").text("关闭授权").click_if_exists(5)

    def release(self):
        logger.info("Native release")
        self.release_auto_authorize()
        if self.perf_flag:
            self.stop_get_perf()
        self.at.release()
        self._empty_base_screen_dir(self.outputs_screen)  # 清空截图目录

    def start_wechat(self):
        if self.at.adb.app_is_running(WECHAT_PACKAGE):
            if self.debug:
                return
            self.at.adb.stop_app(WECHAT_PACKAGE)  # 先关掉
        self.at.adb.start_app(WECHAT_PACKAGE, WECHAT_ACTIVITY)
        # 检测微信画面, 10s
        timeout = time.time() + 10
        while time.time() < timeout:
            views = self.at.java_driver.dump_ui(1)
            texts = set([v.text for v in views])
            if texts.issuperset(WORDING.COMMON.LOGIN_WORDS.zh):
                WORDING.setLanguage(Language.zh)
                logger.info("wechat language: %s" % Language.zh.name)
                break
            elif texts.issuperset(WORDING.COMMON.LOGIN_WORDS.en):
                WORDING.setLanguage(Language.en)
                logger.info("wechat language: %s" % Language.en.name)
                break
        else:
            logger.info("wechat language use default: %s" % WORDING.language.name)
        

    def fold_keybroad(self):
        self.at.adb.press_back()

    def to_main_ui(self, timeout=3):
        el1 = self.e.pkg(WECHAT_PACKAGE).desc("返回")
        el2 = self.e.pkg(WECHAT_PACKAGE).desc("关闭")
        while True:
            try:
                if el1.exists(timeout):
                    el1.click()
                elif el2.exists(timeout):
                    el2.click()
                else:
                    break
            except:
                logger.exception("返回 failed")
                pass

    def stop_wechat(self):
        self.at.adb.stop_app(WECHAT_PACKAGE)

    def click_coordinate(self, x, y):
        self.at.adb.click_point(x, y)

    def click_point(self, x, y):
        """
        Alias for click_coordinate
        """
        return self.click_coordinate(x, y)

    def _get_mem_used_new(self, pkg_name, pid):
        used = 0
        i = 0
        m = None
        m1 = None
        m2 = None
        cmd = "dumpsys meminfo %s" % pkg_name
        # logger.info(cmd)
        output_mem = self.at.adb.run_shell(cmd)
        # logger.info(output_mem)
        r = re.compile(r"TOTAL\s+(\d+)")

        r2 = re.compile(r"size:\s+\S+\s+\S+\s+\S+\s+(\d+)")
        ls = re.split(r"[\r\n]+", output_mem)
        # logger.info(ls)

        for line in ls:
            # i = i + 1
            # if i == 2:
            #     if line.find(pkg_name):
            #         logger.info("line is %s"%line)
            #         m1 = line.split("K")
            m = r.search(line)

            m2 = r2.search(line)

            m1 = line.find(pkg_name)
            # if m1:
            #     search_k = line.find("K:")
            #     search_pid = line.find("pid")
            #     if search_k and search_pid:
            #         num = line.split("K:")[0]
            #         logger.info("num is %s" % num)
            #         used = int(num) / 1024

            if m:
                used = int(m.group(1)) / 1024

            elif m2:
                used = int(m2.group(1)) / 1024

            # elif m1:
            #     logger.info("m1 is %s" % m1)
            #     used = int(m1) / 1024
        if used == 0:
            try:
                used = self.get_mem_used_new_doudi(output_mem, pkg_name, pid)
            except Exception as e:
                logger.error(e)
        return str(used)

    def get_mem_used_new_doudi(self, output_mem, processname, pid):
        used = 0
        if output_mem.find("Total PSS by process"):
            ls = re.split(r"[\r\n]+", output_mem)
            tag = processname + " (pid " + str(pid) + " / activities)"
            for line in ls:
                if line.find(tag) > -1:
                    used_str = line.split("K: ")[0]
                    used = int(used_str.replace(",", "")) / 1024
                    break
        return used

    def _get_cpu_rate_by_file(
        self, mPid, mLastPidUserCpuTime, mLastPidKernelCpuTime, mLastTotalCpuTime
    ):
        # 计算cpu，读取/proc/stat的数据，计算cpuTime
        # 读取/proc/mPid/stat的数据，计算当前小程序的cpu数据
        pid_stat_file = "/proc/" + str(mPid) + "/stat"
        outputs = self.at.adb.run_shell("cat %s" % (pid_stat_file))
        # logger.info(outputs)
        pid_cpu_info_list = outputs.split(" ")
        # logger.info(pid_cpu_info_list)
        if len(pid_cpu_info_list) < 17:
            return 0, mLastPidUserCpuTime, mLastPidKernelCpuTime, mLastTotalCpuTime
        cpuTime = self._get_cpu_time()
        # logger.info(cpuTime)
        # 该进程处于用户态的时间
        pidUTime = int(pid_cpu_info_list[13])
        # logger.info(pidUTime)
        # 该进程处于内核态的时间
        pidSTime = int(pid_cpu_info_list[14])
        # logger.info(pidSTime)
        pidCpuUsage = 0
        # logger.info(mLastPidUserCpuTime)
        if mLastTotalCpuTime != 0:
            pidUserCpuUsage = (
                float(pidUTime - mLastPidUserCpuTime)
                / float(cpuTime - mLastTotalCpuTime)
            ) * 100
            pidKernelCpuUsage = (
                float(pidSTime - mLastPidKernelCpuTime)
                / float(cpuTime - mLastTotalCpuTime)
            ) * 100
            pidUserCpuUsage = max(0, pidUserCpuUsage)
            pidKernelCpuUsage = max(0, pidKernelCpuUsage)
            pidCpuUsage = pidUserCpuUsage + pidKernelCpuUsage

        mLastTotalCpuTime = cpuTime
        mLastPidUserCpuTime = pidUTime
        mLastPidKernelCpuTime = pidSTime
        # logger.info(pidCpuUsage)
        return (
            pidCpuUsage,
            mLastPidUserCpuTime,
            mLastPidKernelCpuTime,
            mLastTotalCpuTime,
        )

    def _get_cpu_time(self):
        stat_file = "/proc/stat"
        outputs = self.at.adb.run_shell("cat %s" % (stat_file))
        # logger.info(outputs)
        lines = outputs.split("\n")
        list = lines[0].split(" ")
        # logger.info(list)
        if len(list) < 9:
            return 0
        user = int(list[2])
        nice = int(list[3])
        system = int(list[4])
        idle = int(list[5])
        ioWait = int(list[6])
        hardIrq = int(list[7])
        softIrq = int(list[8])
        stealTime = int(list[9])

        cpuTime = user + nice + system + idle + ioWait + hardIrq + softIrq + stealTime
        return cpuTime

    def write_perf_data(self, processname, pid, curview, timeinterval):
        # 对齐
        mLastPidUserCpuTime = 0
        mLastPidKernelCpuTime = 0
        mLastTotalCpuTime = 0
        while self.perf_flag:
            stime = time.time()
            # get cpu
            (
                perf_cpu,
                mLastPidUserCpuTime,
                mLastPidKernelCpuTime,
                mLastTotalCpuTime,
            ) = self._get_cpu_rate_by_file(
                pid, mLastPidUserCpuTime, mLastPidKernelCpuTime, mLastTotalCpuTime
            )
            # get fps
            if curview:
                fps_lines = self.at.adb.get_fps(curview)
                if len(fps_lines) <= 8:
                    doudi_currentview = (
                        "com.tencent.mm/com.tencent.mm.plugin.appbrand.ui.AppBrandUI#0"
                    )
                    fps_lines = self.at.adb.get_fps(doudi_currentview)
                try:
                    refresh_period, timestamps1 = self._collectFPSData(fps_lines)
                    fps_data = self._CalculateResults(refresh_period, timestamps1)
                except Exception as e:
                    logger.error(e)
                    # logger.info(fps_lines)
                    fps_data = 0

            else:
                fps_data = 0
            # get mem
            perf_mem = self._get_mem_used_new(processname, pid)

            # record
            logger.info(
                "perf_cpu: %s, fps_data: %s, perf_mem: %s"
                % (perf_cpu, fps_data, perf_mem)
            )
            timestamp = int(time.time())
            self.perf_data.append(
                {
                    "cpu": perf_cpu,
                    "mem": float(perf_mem),
                    "fps": fps_data,
                    "timestamp": timestamp,
                }
            )
            self._screen_shot(os.path.join(self.outputs_screen, "%d.png" % timestamp))
            etime = time.time()
            if etime - stime > timeinterval:
                continue
            time.sleep(etime - stime)

    def _start_get_perf(self, timeinterval=15):
        self.perf_flag = True
        retry = 3
        while retry > 0 and self.perf_flag:
            retry -= 1
            processname, pid = self.at.adb.get_current_active_process(
                re.escape("com.tencent.mm:appbrand") + "\d+"
            )
            logger.info("processname [%s] pid [%s]" % (str(processname), str(pid)))
            self.at.adb.clearbuffer()
            curview = self.at.adb.get_current_active_view(
                re.escape("com.tencent.mm/com.tencent.mm.plugin.appbrand.ui.AppBrandUI")
                + "\d*#\d+"
            )
            if (not processname) or (not curview):
                logger.error("current activity is not appbrand")
                time.sleep(2)
                continue
            else:
                break
        if processname and curview:
            try:
                self.perf_data = []
                self.write_perf_data(processname, pid, curview, timeinterval)
                return self.perf_flag
            except:
                self.perf_flag = False
                return self.perf_flag
        else:
            self.perf_flag = False
            return self.perf_flag

    def start_get_perf(self, timeinterval=15):
        if self.perf_flag:  # 一个线程就够了
            return True
        self.perf_thread = threading.Thread(
            target=self._start_get_perf,
            args=(timeinterval,),
        )
        self.perf_thread.daemon = True
        self.perf_thread.start()
        return True

    def stop_get_perf(self):
        self.perf_flag = False
        if self.perf_thread:
            self.perf_thread.join()
        if self.perf_data:
            perf_data_str = json.dumps(self.perf_data)
            self.perf_data = []
            return perf_data_str
        else:
            logger.error("get perf data fail")
            return ""

    def get_pay_value(self):
        try:
            if self.e.text_contains("￥").exists():
                return self.e.text_contains("￥").get_text()
            else:
                return ""
        except:
            return ""

    def _check_pay_modal_version(self):
        if not self.pay_modal_version:  # 先检测版本
            self._get_current_views()
            if self._el_exist_in_current_view(
                self.e.rid("com.tencent.mm:id/tenpay_keyboard_0")
            ):
                self.pay_modal_version = 2
            elif self._el_exist_in_current_view(
                self.e.text_contains("请输入支付密码")
            ) and self._el_exist_in_current_view(self.e.text_contains("￥")):
                self.pay_modal_version = 1
            else:
                raise TypeError("暂不支持当前微信客户端版本")
        return self.pay_modal_version

    def input_pay_password(self, psw=""):
        check_flag = self.e.text("请输入支付密码").click_if_exists(5)
        if not check_flag:
            return False
        if not psw:
            raise TypeError("未传入支付密码")
        psw = str(psw)
        if type(eval(psw)) is not int:
            raise TypeError("非法支付密码字符串")
        elif len(psw) != 6:
            raise TypeError("密码长度必须为6位数")
        self._check_pay_modal_version()
        for number in psw:
            if self.pay_modal_version == 1:
                point = self._get_number_point(number)
                self.at.device.click_on_point(point["x"], point["y"])
            elif self.pay_modal_version == 2:
                self.e.rid(f"com.tencent.mm:id/tenpay_keyboard_{number}").click()

        if self.e.text("支付成功").exists(10):
            self.e.text("完成").click(is_scroll=True)
        else:
            raise TypeError("支付失败")

    def close_payment_dialog(self):
        if self.e.text_contains("￥").exists(10):
            if not self.e.desc("关闭").click_if_exists():
                # 找不到关闭按钮试下直接返回
                self._press_back()
        else:
            if self.e.text_contains("支付方式").exists():
                self.at.adb.press_back()  # 有可能找不到输入框，先后退尝试
                return True
            logger.error("支付框不存在")
        return False

    def text_exists(self, text="", iscontain=False, wait_seconds=5):
        if iscontain:
            return self.e.text_contains(text=text).exists(wait_seconds)
        else:
            return self.e.text(text=text).exists(wait_seconds)

    def text_click(self, text="", iscontain=False):
        if iscontain:
            return self.e.text_contains(text=text).click(is_scroll=True)
        else:
            return self.e.text(text=text).click(is_scroll=True)

    def get_system_info(self):
        return self.at.adb.desc

    def _press_back(self):
        try:
            self.at.adb.press_key_code(at.keycode.KEYCODE_BACK)
        except:
            return False
        return True

    def is_app_in_foreground(self, appid):
        # exists text like appid:page_path:VISIBLE
        return self.e.text_contains(text=appid).exists(0.5)

    # back_to_miniprogram needed
    def _is_in_wechat(self, activity: str):
        return activity.startswith(WECHAT_PACKAGE)

    def _is_in_wechat_main(self, activity: str):
        return activity.endswith(WECHAT_ACTIVITY)

    def _is_in_miniprogram(self, activity: str):
        # return re.sub(r"\d+$", "", activity).endswith(MINIPROGRAM_ACTIVITY)
        temp = re.sub(r"\d+$", "", activity)
        return temp.endswith(MINIPROGRAM_ACTIVITY) or temp.endswith(
            MINIPROGRAM_PLUGIN_ACTIVITY
        )

    def _is_in_target_miniprogram(self, appid: str):
        if not appid:
            return True
        appids = []
        for i in self.at.java_driver.dump_ui():
            match = re.match("^(wx\w+):\w+$", i.content_desc)
            if match:
                appids.append(match.group(1))
        return appids[-1] == appid if len(appids) > 0 else False

    def _close_miniprogram(self):
        return self._press_back()

    def _get_any_modal(self, confirm=False):
        """
        confirm == True: 确认/允许
        confirm == False: 拒绝/取消
        蒙层下有button可以点击, 可以覆盖:
        1. 授权弹窗 —— 拒绝
        2. 分享弹窗 —— 取消/发送
        3. 模态弹窗 —— 取消/确定
        4. ACTION SHEET暂不支持
        """
        self._get_current_views()  # 刷新一下当前所有的views信息
        modals = []
        buttons = ["允许", "确认", "确定"] if confirm else ["拒绝", "取消"]
        for button in buttons:
            btn = (
                self.e.rid("android:id/content")
                .child()
                .focusable(True)
                .cls_name("android.widget.Button")
                .text(button)
            )
            if self._el_exist_in_current_view(btn):
                modals.append(btn)

        if not modals and not confirm:  # 兜底弹窗
            ok = (
                self.e.rid("android:id/content")
                .child()
                .focusable(True)
                .cls_name("android.widget.Button")
                .text("确定")
            )
            if self._el_exist_in_current_view(ok):
                modals.append(ok)

        return modals

    def _get_current_activity(self):
        output = self.at.adb.run_adb(
            [
                "shell",
                'dumpsys activity activities | grep -E "mFocusedActivity|mResumedActivity"',
            ]
        )
        logger.debug(output)
        for line in output.split("\n"):
            line = line.strip()
            m = re.search(r"((\w+?\.)+?(\w+)/(\.\w+)+)", line)
            if m:
                logger.info(f"current activity is {m.group(0)}")
                return m.group(0)
        return ""

    def _is_in_payment(self):
        return self.e.text_contains("￥").exists(0.5) or self.e.text_contains(
            "支付方式"
        ).exists(0.5)

    def _handle_modal(self, modals):
        if not modals:
            return ModalStatus.OK
        if isinstance(modals, (tuple, list)):
            success_modals = []
            ret = ModalStatus.OK
            for modal in modals:
                logger.info(f"handle modal: {modal}")
                try:
                    rid = modal.get_view().rid
                except at.core.exceptions.UiNotFoundError:
                    return ModalStatus.NotFound
                if modal.click_if_exists(0):
                    if not rid:
                        ret = ModalStatus.NotOfficialModal
                    success_modals.append(modal)
            logger.info(
                f"need handle {len(modals)} modal, success {len(success_modals)}"
            )
            if len(success_modals) > 0:
                return ret
        else:
            modal = modals
            logger.info(f"handle modal: {modal}")
            try:
                rid = modal.get_view().rid
            except at.core.exceptions.UiNotFoundError:
                return ModalStatus.NotFound
            if modal.click_if_exists(0):
                if not rid:
                    return ModalStatus.NotOfficialModal
                return ModalStatus.OK
            return ModalStatus.Error
        return ModalStatus.Error

    def _get_number_point(self, number):
        """
        获取数字的中心坐标
        :param number:
        :return:
        """
        if type(number) is str:
            if type(eval(number)) is not int:
                raise TypeError("number应该为int")
            number = int(number)
        rect = self.e.rid("com.tencent.mm:id/g60").get_rect()
        column_distance = (rect.right - rect.left) / 3  # 列距
        line_distance = (rect.bottom - rect.top) / 4  # 行距

        start_center_point = {
            "x": rect.left + column_distance / 2,
            "y": rect.top + line_distance / 2,
        }

        if number == 1:
            x = start_center_point["x"]
            y = start_center_point["y"]
        elif number == 0:
            x = start_center_point["x"] + column_distance
            y = start_center_point["y"] + line_distance * 3
        else:
            row_num = (number - 1) // 3  # 跟起点相差几行
            column_num = number % 3
            if column_num == 0:
                column_num = 3
            column_num = column_num - 1  # 跟起点相差几列
            x = start_center_point["x"] + column_distance * column_num
            y = start_center_point["y"] + line_distance * row_num
        print("number:{0} ===> Point({1},{2})".format(number, x, y))
        return {"x": x, "y": y}

    def _collectFPSData(self, results):
        timestamps = []
        if len(results) < 128:
            return (0, timestamps)
        nanoseconds_per_second = 1e9
        refresh_period = int(results[0]) / nanoseconds_per_second
        # logger.debug("refresh_period: %s" % refresh_period)
        pending_fence_timestamp = (1 << 63) - 1
        for line in results[1:]:
            fields = line.split()
            if len(fields) != 3:
                continue
            if not fields[0].isnumeric():
                continue
            if int(fields[0]) == 0:
                continue
            timestamp = int(fields[1])
            if timestamp == pending_fence_timestamp:
                continue
            timestamp /= nanoseconds_per_second
            timestamps.append(timestamp)
        # logger.error("timestamps: %s" % timestamps)
        return (refresh_period, timestamps)

    def _GetNormalizedDeltas(self, data, refresh_period, min_normalized_delta=None):
        deltas = [t2 - t1 for t1, t2 in zip(data, data[1:])]
        if min_normalized_delta is not None:
            deltas = filter(
                lambda d: d / refresh_period >= min_normalized_delta, deltas
            )
            deltas = list(deltas)
        return (deltas, [delta / refresh_period for delta in deltas])

    def _CalculateResults(self, refresh_period, timestamps):
        """Returns a list of SurfaceStatsCollector.Result."""
        frame_count = len(timestamps)
        if frame_count == 0:
            return 0
        seconds = timestamps[-1] - timestamps[0]
        frame_lengths, normalized_frame_lengths = self._GetNormalizedDeltas(
            timestamps, refresh_period, _MIN_NORMALIZED_FRAME_LENGTH
        )
        if len(frame_lengths) < frame_count - 1:
            logging.warning("Skipping frame lengths that are too short.")
            frame_count = len(frame_lengths) + 1
        if len(frame_lengths) == 0:
            # raise Exception("No valid frames lengths found.")
            return 0
        return int(round((frame_count - 1) / seconds))

    @property
    def e(self):
        return self.at.e

    # 云环境上莫名其妙卡死在这里了。。。
    @wait(60, False)
    def check_connection(self, *args):
        # 检查atsub链接
        if self.at.java_driver.ping():
            return True
        # reconnect
        try:
            self.at.java_driver.close_remote()
            time.sleep(1)
            self.at.java_driver._init(True)
        except Exception as e:
            logger.exception(e)
            return False
        return self.at.java_driver.is_remote_running()
    
    def close_local_debug_modal(self):
        self.handle_modal(
            btn_text=WORDING.COMMON.MODAL_CONFIRM.value, title=WORDING.COMMON.LOCAL_DEBUG_MODAL_TITLE.value, force_title=True
        ) or self.handle_modal(
            btn_text=WORDING.COMMON.MODAL_CONFIRM.value, title=WORDING.COMMON.LOCAL_DEBUG_MODAL_TITLE2.value, force_title=True
        )


if __name__ == "__main__":
    n = WXAndroidNative({"lang": "en"})
    n.handle_modal("queding", "")



#!/usr/local/bin/python2.7
# encoding: utf-8
'''
Created on 2014-8-28

@author: xiazeng
'''
import datetime
import logging
import os
import os.path
import re
import subprocess
import sys
import tempfile
import threading
import time
import traceback

import at.keycode
import at.utils.apkinfo
from at.utils import decorator, magic

from .exceptions import *

logger = logging.getLogger()

if sys.version_info[0] < 3:
    basestring = (str, unicode)
else:
    basestring = str


def run_cmd(cmd, timeout_sec=60, sync=True):
    """
    不支持重定向命令， 如: echo aa >> output
    """
    def _cb(p):
        logger.error("timeout run cmd: %s", cmd)
        p.kill()

    if not magic.is_windows():
        if isinstance(cmd, basestring):
            args = re.split(r"\s+", cmd)
        elif isinstance(cmd, list):
            args = cmd
        else:
            raise TypeError(u"cmd should be string or list")

        proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    else:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)

    if not sync:
        return proc
    timer = threading.Timer(timeout_sec, _cb, [proc])
    try:
        timer.start()
        return proc.communicate()
    finally:
        timer.cancel()


def get_device_ids():
    f = os.popen("adb devices")
    devices = f.read().strip().split('\n')
    f.close()
    devices = devices[1:]
    ret = []
    for device in devices:
        serial = re.split(r'\s+', device)[0].strip()
        status = re.split(r'\s+', device)[1].strip()
        if status == 'device':
            ret.append(serial)
    return ret


def get_device_name(device_id):
    prop = os.popen('adb -s '+device_id+' shell getprop').read()
    rets = re.findall(r"\[ro.product.model\]: \[(.*)\]", prop)
    device_name = ''
    if rets:
        device_name = rets[0]
        device_name = device_name.replace(' ', '_')
    return device_name


def get_devices_names():
    ret = {}
    for device in get_device_ids():
        ret[device] = get_device_name(device)
    return ret


def get_key_value(s, key):
    r = re.compile(key+r"\s*=\s*(.*)$", re.MULTILINE)
    m = r.search(s)
    if m:
        return m.group(1)
    return None


class AdbDumpSys(object):
    def __init__(self, serial):
        self._adb = AdbWrap.apply_adb(serial)

    def _cmd(self, args):
        return self._adb.run_shell("dumpsys " + args)

    def package(self, pkg_name):
        output = self._cmd("package " + pkg_name)

        class PackageInfo(object):
            def __init__(self, s):
                self.versionName = get_key_value(s, "versionName")
                self.firstInstallTime = get_key_value(s, "firstInstallTime")
                self.lastUpdateTime = get_key_value(s, "lastUpdateTime")

            def is_first_install(self):
                return self.firstInstallTime == self.lastUpdateTime
        if "Unable to find package" in output:
            # logger.debug(output)
            return None
        return PackageInfo(output)


class AdbWrap(object):
    _serial_kv = {}
    adb_helper_dict = {}
    error_tags = [
        "no devices/emulators found",
        "error: closed",
        "more than one device/emulator",
        "device offline",
        "device not found",
        "protocol fault"
    ]

    def __init__(self, serial=None):
        self.adb_path = "adb"
        self._sdcard_path = None
        self._serial = serial
        self._last_pid = None
        self._error_msg = None
        self._last_output = None
        self._is_connected = True
        self._disconnected_cb = None
        self._call_back_list = []
        self._has_call_back = False
        serial_ids = get_device_ids()
        logger.info(str(serial_ids))
        if serial is None:
            self._serial = self.get_default_serial()
        if serial is not None and serial not in serial_ids:
            logger.error("可用手机:%s, '%s'不可用", serial_ids, serial)
            raise AdbDisconnectedError(u"手机可用:"+serial)
        if self._serial is None and self.kv is None:
            self._serial_kv['none'] = {}
        elif self._serial is not None and self.kv is None:
            self._serial_kv[self._serial] = {}

    def check_connected(self):
        self.run_adb("shell echo 1")
        return self._is_connected

    def is_connected(self):
        return self._is_connected

    def add_disconnected_cb(self, _cb):
        self._call_back_list.append(_cb)
        self.check_connected()

    def set_connected_state(self, state):
        if not state:
            if not self._has_call_back:
                self._has_call_back = True
                for _cb in self._call_back_list:
                    _cb(self.serial, self._last_output, self._error_msg)

        self._is_connected = state

    def is_locked(self):
        output = self.run_shell(" dumpsys window policy")
        for line in output.split("\n"):
            if "isStatusBarKeyguard" in line and "true" in line:
                return True
            if "mShowingLockscreen" in line and "true" in line:
                return True
        return False

    @classmethod
    def get_default_serial(cls):
        serial_ids = get_device_ids()
        if len(serial_ids) == 0:
            raise AdbException(u"没有在线的手机")
        elif len(serial_ids) > 1:
            raise AdbException(u"多台%d手机在线，请指定serial, serials:%s" % (len(serial_ids), serial_ids))
        else:
            return serial_ids[0]

    @classmethod
    def apply_adb(cls, serial=None):
        if serial is None:
            serial = cls.get_default_serial()
        # logger.debug('serial is: %s, %s', serial, cls.adb_helper_dict.keys())
        if serial in cls.adb_helper_dict:
            # logger.debug("use cache adbHelper: " + str(serial))
            return cls.adb_helper_dict[serial]
        else:
            # logger.debug("create adbHelper: " + str(serial))
            helper = AdbWrap(serial)
            cls.adb_helper_dict[serial] = helper
            return helper

    def _run(self, cmd, sync=True, timeout_sec=60, retry=3):
        """
        :rtype: str
        """
        logger.debug(cmd)
        if not sync:
            return run_cmd(cmd, timeout_sec, sync)
        output, error = run_cmd(cmd, timeout_sec, sync)
        try:
            if magic.is_windows():
                error = error.decode("gbk")
                output = output.decode("gbk")
            else:
                error = error.decode("utf-8")
                output = output.decode("utf-8")
        except UnicodeDecodeError:
            pass

        self._last_output = output
        self._error_msg = error
        if isinstance(output, bytes):
            return output
        for error_tag in self.error_tags:

            if (error and error_tag in error) or error_tag in output:
                logger.error(error)
                if retry > 0:
                    time.sleep(11)
                    logger.warning("10 seconds later try again")
                    return self._run(cmd, sync, timeout_sec, retry-1)
                else:
                    self.set_connected_state(False)
        if error:
            logger.warning(error)

        return output

    def post_run(self, cmd, sync=True, timeout_sec=60):
        logger.debug(cmd)
        return run_cmd(cmd, timeout_sec, sync)

    def get_output_text(self):
        return self._last_output

    def get_output_lines(self):
        if self.get_output_text():
            lines = self.get_output_text().split("\n")
            return [l.strip() for l in lines]
        else:
            return []

    def get_error(self):
        return self._error_msg

    def clear_error(self):
        self._error_msg = None

    def set_error(self, msg):
        self._error_msg = msg

    @property
    def serial(self):
        return self._serial

    def restart_adb(self):
        self.run_adb("kill-server")
        self.run_adb("start-server")

    @property
    def kv(self):
        if self._serial is None:
            return self._serial_kv['none'] if "none" in self._serial_kv else None
        else:
            return self._serial_kv[self._serial] if self._serial in self._serial_kv else None

    def get(self, k):
        value = self.kv[k] if k in self.kv else None
        logger.debug("get in %s, %s=%s" % (str(self._serial), str(k), str(value)))
        return value

    def set(self, k, v):
        logger.debug("set in %s, %s=%s" % (str(self._serial), str(k), str(v)))
        self.kv[k] = v

    def unset(self, k):
        logger.debug("unset in %s, %s" % (str(self._serial), str(k)))
        if k in self.kv:
            del self.kv[k]

    def last_pid(self):
        return self._last_pid

    def prefix(self):
        if self._serial is not None:
            return "%s -s %s " % (self.adb_path, self._serial)
        else:
            return "%s " % self.adb_path

    def is_pid_running(self, pid):
        """
        判断pc的进程是否在运行
        :param pid: pc进程的process id
        :rtype: bool
        """
        if magic.is_windows():
            cmd = "tasklist /FI \"PID eq %d\"" % pid
            output = self._run(cmd)
            if str(pid) in output:
                return True
            else:
                return False
        else:
            try:
                os.kill(pid, 0)
            except OSError:
                return False
            else:
                return True

    @at.utils.decorator.retry_in(5, 1)
    def android_pid_running(self, pid, debug=False):
        cmds = ["ps", "ps -A"]
        outputs = []
        for cmd in cmds:
            output = self.run_shell(cmd)
            logger.warning(output)
            lines = output.split("\n")
            for line in lines[1:]:
                line = line.strip()
                ls = re.split(r"\s+", line)
                if len(ls) < 9:
                    continue
                if pid == int(ls[1]):
                    return True
            outputs.append(output)
        logger.warning("pid: %s not exists", pid)
        if debug:
            for output in outputs:
                logger.info(output)
        return False

    def get_android_pid(self, pkgname):
        cmds = ["ps", "ps -A"]
        for cmd in cmds:
            output = self.run_shell(cmd)
            lines = output.split("\n")
            for line in lines[1:]:
                line = line.strip()
                ls = re.split(r"\s+", line)
                if len(ls) < 9:
                    continue
                if pkgname == ls[8]:
                    return int(ls[1])
        return None

    def get_all_android_pids(self, reg=None):
        pid_mapping = {}
        cmds = ["ps", "ps -A"]
        for cmd in cmds:
            output = self.run_shell(cmd)
            lines = output.split("\n")
            for line in lines[1:]:
                line = line.strip()
                ls = re.split(r"\s+", line)
                if len(ls) < 9:
                    continue
                elif not reg or reg.match(ls[8]):
                    pid_mapping[ls[8]] = int(ls[1])
        return pid_mapping

    def get_android_uid(self, pkgname):
        try:
            pkg_pid = self.get_android_pid(pkgname)
            status_filename = "/proc/%d/status" % pkg_pid
            output = self.run_shell("cat %s" % status_filename)
            m = re.search(r"[Uu]id:\s+(\d+)", output, re.S)
            return int(m.group(1))
        except Exception as e:
            traceback.print_exc()
            return None

    def get_device_id(self):
        hex_id = self.run_shell("settings get secure android_id").strip()
        return int("0x{}".format(hex_id), 16) % 2147483647

    def get_brigthness(self):
        right = self.run_shell("dumpsys power").strip()
        return right

    def get_screen(self):
        screen = self.run_shell("dumpsys window policy").strip()
        return screen
    """
    流量统计相关，参考：http://stackoverflow.com/questions/12904809/tracking-an-applications-network-statistics-netstats-using-adb
    """
    def get_total_traffic_stats(self):
        output = self.run_shell("cat /proc/1/net/dev")
        lines = output.split("\n")
        m_tx = m_rx = wifi_tx = wifi_rx = 0
        for line in lines[2:]:
            line = line.strip()
            ls = re.split(r"\s+", line)
            if ls[0].startswith("rmnet") or ls[0].startswith("ccmni"):
                m_tx += int(ls[9])
                m_rx += int(ls[1])
            elif "lo" != ls[0]:
                wifi_tx += int(ls[9])
                wifi_rx += int(ls[1])
        return m_rx,m_tx, wifi_rx, wifi_tx

    def get_pkg_traffic_stats(self, pkgname):
        uid = self.get_android_uid(pkgname)
        output = self.run_shell("cat /proc/net/xt_qtaguid/stats")
        lines = output.split("\n")
        m_tx = m_rx = wifi_tx = wifi_rx = 0
        for line in lines[1:]:
            line = line.strip()
            ls = re.split(r"\s+", line)
            if len(ls)>3 and int(ls[3]) == uid:
                if ls[1].startswith("rmnet") or ls[1].startswith("ccmni"):
                    m_rx += int(ls[5])
                    m_tx += int(ls[7])
                elif ls[1] != "lo":
                    wifi_rx += int(ls[5])
                    wifi_tx += int(ls[7])
        return m_rx, m_tx, wifi_rx, wifi_tx

    def get_input_methods(self):
        outputs = self.run_shell("ime list -s")
        return [line.strip() for line in outputs.split("\n")]

    def set_input_method(self, input_method):
        return self.run_shell("ime set %s" % input_method)

    @decorator.cached_property
    def screen_size(self):
        output = self.run_shell("wm size")
        m = re.search(r"(\d+)x(\d+)", output)
        if m is not None:
            return int(m.group(1)), int(m.group(2))
        else:
            return 0, 0

    @decorator.cached_property
    def language(self):
        return self.get_property("persist.sys.language")

    @decorator.cached_property
    def sdk_version(self):
        version = self.get_property("ro.build.version.sdk")
        ret = -1
        try:
            ret = int(version)
        except Exception:
            pass
        return ret

    @decorator.cached_property
    def model(self):
        return self.get_property("ro.product.model")

    @decorator.cached_property
    def brand(self):
        return self.get_property("ro.product.brand")

    @decorator.cached_property
    def display(self):
        return self.get_property("ro.build.display.id")

    @decorator.cached_property
    def release(self):
        return self.get_property("ro.build.version.release")

    @decorator.cached_property
    def name(self):
        return self.brand

    @decorator.cached_property
    def desc(self):
        return {
            "name": self.brand,
            "model": self.model,
            "cpu": self.cpu_num,
            "language": self.language,
            "width": self.screen_size[0],
            "height": self.screen_size[1],
            "release": self.release,
            "sdk_version": self.sdk_version,
            "get_mem_info": self.get_mem_info(),
            # "display": self.display,
            "serial": self.serial
        }

    @decorator.cached_property
    def cpu_num(self):
        output = self.run_shell("ls /sys/devices/system/cpu/")
        m = re.findall(r"cpu[0-9]", output, re.M)
        return len(m)

    def get_property(self, name):
        return self.run_shell("getprop "+name).strip()

    def kill_pid(self, pid):
        if not self.is_pid_running(pid):
            return
        if magic.is_windows():
            cmd = "taskkill /F /PID %d" % pid
        else:
            cmd = "kill -9 %d" % pid
        self._run(cmd)

    def kill_by_name(self, name):
        if magic.is_windows():
            cmd = "TASKKILL /F /IM %s " % name
        else:
            cmd = "killall %s" % name
        logger.info(cmd)
        self._run(cmd)

    """
    android 定义了很多keycode，如果要增加API请参考：
    https://android.googlesource.com/platform/frameworks/base/+/master/core/java/android/view/KeyEvent.java
    """
    def press_key_code(self, code):
        self.run_shell("input keyevent %d" % code)

    def press_back(self):
        self.press_key_code(at.keycode.KEYCODE_BACK)
        time.sleep(2)

    def press_enter(self):
        logger.debug("press enter")
        self.press_key_code(at.keycode.KEYCODE_ENTER)

    def press_search(self):
        self.press_key_code(at.keycode.KEYCODE_SEARCH)

    def press_delete(self, num=1):
        for i in range(num):
            self.press_key_code(at.keycode.KEYCODE_DEL)

    def press_menu(self):
        self.press_key_code(at.keycode.KEYCODE_MENU)

    def press_home(self):
        self.press_key_code(at.keycode.KEYCODE_HOME)

    def press_app_switch(self):
        self.press_key_code(at.keycode.KEYCODE_APP_SWITCH)

    def press_number(self, number):
        number = str(number)
        for i in number:
            try:
                self.run_shell("input keyevent %d" % (int(i) + 7))
            except ValueError:
                logger.error(u"number must be string")
                break

    def click_point(self, x, y):
        logger.info("click %d, %d" % (x, y))
        self.run_shell("input tap %d %d " % (x, y))
        time.sleep(0.5)

    def click_point_for_quick(self, x, y):
        logger.info("click %d, %d" % (x, y))
        click_time = self.run_shel_for_quick("input tap %d %d " % (x, y))
        logger.info(click_time)
        return click_time
        # time.sleep(0.5)

    @at.utils.decorator.retry_in(10, 1)
    def get_current_activity(self):
        """
        :return: activity名称，包含包名，如：com.android.launcher/com.android.launcher2.Launcher 或者 com.tencent.mm/.ui.LauncherUI
        """
        act = self.get_current_activity2()
        if not act:
            act = self.get_act_by_dump_act()
        logger.debug('activity: %s', act)
        return act

    def _wrap_activity_name(self, pkg, act):
        """
        统一activity的格式
        :param pkg: 包名，如com.tencent.mm
        :param act: 包含包名，如com.tencent.mm/.ui.LauncherUI
        :return:
        """
        pkg, act = pkg.strip(), act.strip()
        if act.startswith("."):
            return pkg, pkg + '/' + act
        if "/" in act:
            return pkg, act
        return pkg, act

    def _get_top_window(self, lines):
        for line in lines:
            m = re.search(r'mFocusedApp=.* ((\S*)/(.*)) ', line)
            if m:
                return self._wrap_activity_name(m.group(2), m.group(3))

        for line in lines:
           m = re.search(r'mCurrentFocus=.* ((\S*)/(.*?))[ }]', line)
           if m:
               return self._wrap_activity_name(m.group(2), m.group(3))
        return None, None

    def get_act_by_dump_act(self):
        act = None
        output = self.run_shell("dumpsys activity activities")
        try:
            output1 = bytes.decode(output).split("\n")
        except:
            output1=output.split("\n")
        for line in output1:
            line = line.strip()
            if "mFocusedActivity" in line:
                m = re.search(r"((\w+?\.)+?(\w+)/(\.\w+)+)", line)
                if m:
                    act = m.group(0)
                    break
            if "mResumedActivity" in line:
                m = re.search(r"((\w+?\.)+?(\w+)/(\.\w+)+)", line)
                if m:
                    act = m.group(0)
                    break
        return act

    def get_current_activity2(self):
        # as a example:
        # mCurrentFocus=Window{41d2c970 u0 com.android.launcher/com.android.launcher2.Launcher}
        # mFocusedApp=AppWindowToken{4203c170 token=Token{41b77280 ActivityRecord{41b77a28 u0 com.android.launcher/com.android.launcher2.Launcher t3}}}
        output = self.run_adb("shell dumpsys window windows") # | grep -E 'mCurrentFocus|mFocusedApp'
        lines = [line.strip() for line in output.split('\n')]
        _, act = self._get_top_window(lines)
        # logger.debug('activity: %s', act)
        return act

    def get_current_active_process(self, reg_exp):
        """
        grep top 5 process, and return process match {reg_exp}
        :return: process_name, pid
        """
        output = self.run_shell("COLUMNS=512 top -m 5 -n 2 -d 2")
        lines = [line.strip() for line in output.split("\n")]
        for output2 in lines:
            r = re.compile("(%s)" % reg_exp)
            m = r.search(output2)
            if m:
                pid = output2.split()[0]
                m_name = m.group(1)
                return (m_name, pid)
        return (None, None)

    def get_current_active_view(self, reg_exp):
        """
        get current active view from SurfaceFlinger, return view name match {reg_exp}
        """
        output = self.run_adb("shell dumpsys SurfaceFlinger --list")
        lines = [line.strip() for line in output.split("\n")]
        view = None
        r = re.compile("(%s)" % reg_exp)
        for line in lines:
            m = r.search(line)
            if m:
                return m.group(1)
        return None

    def get_current_process(self):
        output = self.run_adb("shell dumpsys window windows")  # | grep -E 'mCurrentFocus|mFocusedApp'
        lines = [line.strip() for line in output.split('\n')]
        pkg, _ = self._get_top_window(lines)
        return pkg

    def get_top_process(self):
        output = self.run_shell("dumpsys activity")
        lines = [line.strip() for line in output.split('\n')]
        for line in lines:
            if "top-activity" in line:
                m = re.search(r"(\d+):(\S+)/(\w+)", line)
                return int(m.group(1)), m.group(2)
        return None, None

    def forward(self, dst_port, src_port):
        cmd = "forward tcp:%s tcp:%s" % (str(dst_port), str(src_port))
        self.run_adb(cmd)

    def forward_name(self, dst_name, src_name):
        cmd = "forward %s %s" % (dst_name, src_name)
        self.run_adb(cmd)

    def forward_remove(self, port):
        cmd = "forward --remove tcp:%s" % str(port)
        self.run_adb(cmd)

    def forward_remove_all(self):
        cmd = "forward --remove-all"
        self.run_adb(cmd)

    def forward_list(self):
        cmd = "adb forward --list"
        lines = self.run_adb(cmd).split("\n")
        ports = []
        for line in lines:
            m = re.match(r".*\s+(\S+):(\S+)\s+(\S+):(\S+)", line)
            if m:
                ports.append()

        self.run_adb(cmd)

    def app_is_running(self, pkg_name):
        pid = self.get_android_pid(pkg_name)
        if pid:
            return True
        else:
            return False

    def app_reset(self, packagename):
        '''
        清空应用数据，慎用
        '''
        cmd = "pm clear %s" % packagename
        return self.run_shell(cmd)

    def pkg_has_installed(self, pkg_name):
        lines = self.run_shell("pm list packages").split('\n')
        for line in lines:
            ls = line.split(":")
            if len(ls) == 2 and pkg_name == ls[1].strip():
                return True
        else:
            return False

    def stop_app(self, pkg_name):
        return self.run_shell("am force-stop %s" % pkg_name)

    def start_app(self, pkg_name, cls_name, option="-W -n", extra=""):
        '''
        :param extra: 如果传入extra参数，那么option必须包含"-W -n"
        '''
        if cls_name.startswith("com"):
            outputs = self.run_shell("am start %s %s/%s %s" % (option, pkg_name, cls_name, extra))
        else:
            if not cls_name.startswith("."):
                cls_name = "." + cls_name
            outputs = self.run_shell("am start %s %s/%s %s" % (option, pkg_name, cls_name, extra))
        logger.debug(outputs)
        lines = outputs.split("\n")

        for line in lines:
            logger.info(line)
            if "Status: ok" in line or "as been brought to the front" in line or "intent has been delivered to currently running top-most instance" in line:
                return True
        else:
            return False

    def restart_app(self, pkg_name, cls_name):
        self.stop_app(pkg_name)
        time.sleep(2)
        self.start_app(pkg_name, cls_name)

    def start_service(self, pkg_name, cls_name):
        if pkg_name.startswith(cls_name):
            outputs = self.run_shell("am startservice --user 0  %s/%s" % (pkg_name, cls_name))
        else:
            if not cls_name.startswith("."):
                cls_name = "." + cls_name
            outputs = self.run_shell("am startservice --user 0 %s/%s" % (pkg_name, cls_name))
        logger.debug(outputs)
        lines = outputs.split("\n")
        for line in lines:
            if "Status: ok" in line:
                return True
        else:
            return False

    def open_activity(self, activity):
        return self.run_shell("am start %s"%activity)

    def get_cpu_rate(self, *pkg_names):
        output = self.run_shell("dumpsys cpuinfo")
        ls = re.split(r"[\r\n]+", output)
        r = re.compile(r'([\.\d]+)%')
        cmd_line_reg = r".*?(?P<all>[\.\d]+)%\s+(?P<pid>\d+)/(?P<name>.*)?:\s+(?P<user>[\.\d]+)%\s+user.*?(?P<kernel>[\.\d]+)%\s+kernel.*"
        ret = {}
        for line in ls:
            for pkg_name in pkg_names:
                if pkg_name in line:
                    m = re.match(cmd_line_reg, line)
                    if m and m.group('name') == pkg_name:
                        ret[pkg_name] = float(m.group(1))
        return ret

    def get_mem_used(self, *pkg_names):
        ret = {}
        for pkg_name in pkg_names:
            output = self.run_shell("dumpsys meminfo %s" % pkg_name)
            r = re.compile(r'TOTAL\s+(\d+)')
            r2 = re.compile(r"size:\s+\S+\s+\S+\s+\S+\s+(\d+)")
            ls = re.split(r"[\r\n]+", output)
            used = 0
            for line in ls:
                m = r.search(line)
                m2 = r2.search(line)
                if m:
                    used = int(m.group(1))/1024
                elif m2:
                    used = int(m2.group(1))/1024
            ret[pkg_name]= used
        return ret

    def get_mem_used_native(self, *pkg_names):
        ret = {}
        for pkg_name in pkg_names:
            output = self.run_shell("dumpsys meminfo %s" % pkg_name)
            r1 = re.compile(r'TOTAL\s+(\d+)')
            r2 = re.compile(r"size:\s+\S+\s+\S+\s+\S+\s+(\d+)")
            r3 = re.compile(r'Native Heap\s+(\d+)') #Native Heap
            r4 = re.compile(r'Java Heap:\s+(\d+)')  #Dalvik Heap
            r5 = re.compile(r'Graphics:\s+(\d+)')  # Graphics:反应webview
            r7 = re.compile(r'Private Other:\s+(\d+)')  # Private Other:其他未分类
            r6 = re.compile(r'Code:\s+(\d+)')  # code：dex、so之类
            ls = re.split(r"[\r\n]+", output)
            used = 0
            native_used = 0
            dalvik_used = 0
            graphics_used = 0
            code_used = 0
            other_used = 0
            for line in ls:
                m1 = r1.search(line)
                m2 = r2.search(line)
                m3 = r3.search(line)
                m4 = r4.search(line)
                m5 = r5.search(line)
                m6 = r6.search(line)
                m7 = r7.search(line)
                if m1:
                    used = int(m1.group(1))/1024
                elif m2:
                    used = int(m2.group(1))/1024
                if m3:
                    native_used = int(m3.group(1))/1024
                if m4:
                    dalvik_used = int(m4.group(1))/1024
                if m5:
                    graphics_used = int(m5.group(1))/1024
                if m6:
                    code_used = int(m6.group(1))/1024
                if m7:
                    other_used = int(m7.group(1))/1024

            ret[pkg_name] = []
            ret[pkg_name].append(used)
            ret[pkg_name].append(native_used)
            ret[pkg_name].append(dalvik_used)
            ret[pkg_name].append(graphics_used)
            ret[pkg_name].append(code_used)
            ret[pkg_name].append(other_used)
        return ret

    def get_vmsize(self, *pkg_names):
        ret = {}
        ret_before = {}
        for pkg_name in pkg_names:

            pid = self.get_android_pid(pkg_name)
            if pid:
                output = self.run_shell("cat /proc/%s/status" % (pid))
                r = re.compile(r'VmSize:\s+(\d+)')
                ls = re.split(r"[\r\n]+", output)
                vmsize =0
                for line in ls:
                    m = r.search(line)
                    if m:
                        vmsize = int(m.group(1))/1024
                ret[pkg_name] = vmsize
        return ret

    def get_fps_rate(self, *pkg_names):
        """
        获取fps
        :return:
        """
        j = 0
        fps_data_list = {}
        fps_data = 0
        output = self.run_shell("dumpsys SurfaceFlinger --list | grep \"com.tencent.mm/com.tencent.mm.plugin.appbrand.ui.AppBrandUI\"") #获取当前页面

        ls = re.split(r"[\r\n]+", output)
        # logger.info(ls)
        current_windows = "com.tencent.mm/com.tencent.mm.plugin.appbrand.ui.AppBrandUI#0"
        # logger.info(current_windows)
        get_data = self.run_shell("dumpsys SurfaceFlinger --latency %s"%(current_windows))
        # logger.info(get_data)
        if(len(get_data.split()) >1):
            get_result = get_data.split()
            # logger.info(get_result)
            end_time = get_result[-2]
            start_time = get_result[2]
            cost_time = int(end_time)-int(start_time)
            # logger.info(cost_time)
            if not cost_time ==0:
                fps_data =126000000000/cost_time
            # logger.info(fps_data)
            # logger.info(get_result[-2])
            # logger.info(get_result[2])
       # for pkg_name in pkg_names:
            fps_data_list["com.tencent.mm:appbrand0"]= fps_data
            fps_data_list["com.tencent.mm:appbrand1"] = 0
            #logger.info(fps_data_list)
            # for i in get_data:
            #     logger.info(i)
            #     j = j + 1
            #     if j==2:
            #         start_time = i.split()[1]
            #         logger.info("start time %s" %(start_time))
            #     if j==128:
            #         end_time = i.split()[1]
            #         logger.info("end_time %s" % (end_time))
            self.run_shell("dumpsys SurfaceFlinger --latency-clear")
        return fps_data_list

    def clearbuffer(self):
        output = self.run_adb("shell dumpsys SurfaceFlinger --latency-clear")
        if not output:
            return True
        else:
            return None

    def get_fps(self, currentview=""):
        output = self.run_adb("shell dumpsys SurfaceFlinger --latency " + currentview)
        lines = [line.strip() for line in output.strip().split("\n")]
        return lines

    def get_mem_info(self):
        output = self.run_shell("cat /proc/meminfo")
        m_total = re.search(r"MemTotal:\s+(\d+)\s*(\w+)", output)
        m_free = re.search(r"MemFree:\s+(\d+)\s*(\w+)", output)
        max_mem = 0
        free_mem = 0
        if None != m_total:
            max_mem = int(m_total.group(1))
            unit = m_total.group(2)
            if unit.upper() == "KB":
                max_mem /= 1024
        if None != m_free:
            free_mem = int(m_free.group(1))
            unit = m_free.group(2)
            if unit.upper() == "KB":
                free_mem /= 1024
        return max_mem, free_mem

    def ping(self, ip, num=10):
        outputs = self.run_shell("ping -c %d %s" % (num, ip))
        logger.debug(outputs)
        lines = outputs.split("\n")
        r = "(?P<num>\d+) packets.*(?P<received>\d+)\s+received, .*time (?P<costs>\d+)ms"
        for l in lines:
            m = re.match(r, l)
            if m:
                return int(m.group("num")), int(m.group("received")), int(m.group("costs"))
        return -1, -1, -1

    # ======================================================================================================
    # 底层命令
    # ======================================================================================================
    def run_shell(self, cmd, sync=True, timeout_sec=30):
        cmd = " shell " + cmd

        return self.run_adb(cmd, sync, timeout_sec=timeout_sec)

    def run_shel_for_quick(self, cmd, sync=True, timeout_sec=30):
        cmd = " shell " + cmd

        click_time = self.run_adb_for_quick(cmd, sync, timeout_sec=timeout_sec)
        logger.info(click_time)
        return click_time

    def broadcast(self, action, params=None):
        if params is None:
            params = {}
        s = []
        for k, v in params.items():
            if isinstance(v, int):
                s.append("--ei %s %s" % (k, v))
            else:
                s.append("-e %s %s" % (k, v))
        self.run_shell("am broadcast -a %s %s" % (action, ' '.join(s)))

    def wait_connect(self, timeout=60):
        """
        等待设备在线
        :param timeout:
        :return:
        """
        t = time.time()
        while time.time() - t < timeout:
            if self._serial in get_device_ids():
                return True
            time.sleep(1)

    def run_adb(self, cmd, sync=True, timeout_sec=30):
        """
        运行命令: adb xxxxx
        - cmd: str || list
        """
        if not self._is_connected:
            raise AdbDisconnectedError("adb has disconnected")
        if isinstance(cmd, list):
            _cmd = []
            if magic.is_windows():
                for c in cmd:
                    _cmd.append(re.sub(r"([\|\^&])", lambda m: "^" + m.group(0) , c))  # |^&在windows中属于特殊字符，不能直接传参数
            else:
                _cmd = cmd
            cmd = re.split(r"\s+", self.prefix().strip()) + _cmd
        else:
            cmd = self.prefix() + (re.sub(r"([\|\^&])", lambda m: "^" + m.group(0) , cmd) if magic.is_windows() else cmd)  # |^&在windows中属于特殊字符，不能直接传参数
        return self._run(cmd, sync, timeout_sec)

    def run_adb_for_quick(self, cmd, sync=True, timeout_sec=30):
        """
        运行命令: adb xxxxx
        - cmd: str || list
        """

        cmd = self.prefix() + cmd
        t = time.time()
        logger.info(t)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        tt = time.time()
        logger.info(tt)

        def _cb(p):
            logger.error("timeout %s", cmd)
            p.kill()
        timer = threading.Timer(timeout_sec, _cb, [proc])
        try:
            timer.start()
            return proc.communicate()
        finally:
            timer.cancel()
            return  tt

    def get_current_processes(self):
        outputs = self.run_shell("ps")
        processes = {}
        lines = outputs.replace("\r\n", "\n").split("\n")
        r = re.compile(r"\s+")
        for line in lines[1:]:
            line = line.strip()
            if line:
                ls = r.split(line)
                pid = int(ls[1])
                process = {
                    "pid": pid,
                    "ppid": int(ls[2]),
                    "name": ls[-1]
                }
                processes[pid] = process
        return processes

    def push(self, local, remote):
        if not os.path.exists(local):
            raise AtError("no local file:"+local)
        cmd = "push %s %s" % (local, remote)
        outputs = self.run_adb(cmd)
        if "failed" in outputs:
            return False
        else:
            return True

    def compare_push(self, path, path_in_android):
        if not os.path.exists(path):
            raise AtError("no local file:"+path)
        if self.file_exists(path_in_android):
            local_minitouch_size = int(os.path.getsize(path))
            file_size = self.file_size(path_in_android)
            if local_minitouch_size == file_size:
                logger.debug("exists skipped")
                return False
        self.push(path, path_in_android)
        return True

    def delete(self, remote, opt=""):
        # self.assert_file_exists(remote)
        cmd = "shell rm %s %s" % (opt, remote)
        return self.run_adb(cmd)

    def pull(self, remote, local):
        if local is None:
            local = os.getcwd()
        parent_directory = os.path.dirname(local)
        if not os.path.exists(parent_directory) and parent_directory.strip():
            os.makedirs(parent_directory)
        if not self.file_exists(remote):
            return False
        else:
            cmd = "pull %s %s" % (remote, local)
            outputs = self.run_adb(cmd)
            if "failed" in outputs or "error" in outputs:
                return False
            return True

    def get_sdcard_path(self):
        if self._sdcard_path is not None:
            return self._sdcard_path
        if self._serial == "f47bd83d":
            self._sdcard_path = "/storage/emulated/0"
        else:
            output = self.run_shell(" echo $EXTERNAL_STORAGE")
            if output is not None and output.strip():
                self._sdcard_path = output.strip()
            else:
                self._sdcard_path = "/sdcard"
        logger.info("sdcard:"+self._sdcard_path)
        return self._sdcard_path

    def get_file_modifytime(self, remote, is_directory=False):
        self.assert_file_exists(remote)
        if is_directory:
            opt = "-ld"
        else:
            opt = "-l"
        output = self.run_shell("ls %s %s" % (opt, remote))
        ls = output.split("\n")
        if len(ls) > 1:
            raise AdbException("%s is directry"%remote)
        m = re.search(r"\d+\-\d+\-\d+\s+\d+:\d+", output)
        if m is None:
            raise AdbException("unknow output:"+output)
        return m.group()

    def screen_cap(self, filename):
        screen_path = self.get_sdcard_path() + "/mmtest/screenshot"
        cmd = "mkdir -p "+screen_path
        self.run_shell(cmd)
        filename = os.path.basename(filename)
        filename = screen_path + "/" + filename
        cmd = "screencap -p "+filename
        self.run_shell(cmd)
        return filename

    def screen_record(self, filename, max_time, sync=True):
        screen_path = self.get_sdcard_path() + "/mmtest/record"
        cmd = "mkdir -p "+screen_path
        self.run_shell(cmd)
        filename = screen_path +"/" + filename
        cmd = "screenrecord --time-limit %s %s"%(max_time, filename)
        self.run_shell(cmd, sync)
        return filename

    @at.utils.decorator.cost_log
    def install(self, filename, opt="-r", uninstall_when_imcompatible=True):
        apk_info = at.utils.apkinfo.ApkInfo(filename)
        dumpsys = AdbDumpSys(self.serial)
        package_info = dumpsys.package(apk_info.pkg)
        # filename 路径上可能存在空格，不能直接使用string的形式
        # cmd = " install %s %s" % (opt, filename)
        cmd = ["install", *re.split(r"\s+", opt), filename]
        output = self.run_adb(cmd, True, 5*60)
        output += str(self._error_msg)
        if "Success" not in output:
            if "INSTALL_FAILED_UPDATE_INCOMPATIBLE" in output and uninstall_when_imcompatible:
                self.uninstall(apk_info.pkg)
                output = self.run_adb(cmd, True, 180)
                output += str(self._error_msg)
                if "Success" not in output:
                    if "Could not access the Package Manager" in output:
                        raise CanNotAccessToPackageManager(output)
                    logger.error(output)
                    return False
            else:
                logger.error(output)
                return False
        s = time.time()
        while time.time() - s < 300:
            current_package_info = dumpsys.package(apk_info.pkg)
            if package_info is None:
                if current_package_info is not None and current_package_info.is_first_install():
                    logger.info("install success, first install")
                    return True
            else:
                if current_package_info is not None:
                    if current_package_info.lastUpdateTime != package_info.lastUpdateTime:
                        return True
            time.sleep(1)
        logger.info("install timeout...")
        return False

    def push_install(self, filename, option="-r"):
        name = os.path.basename(filename)
        remote_path = "/sdcard/%s" % name
        self.push(filename, remote_path)
        output = self.run_shell("pm install %s %s" % (option, remote_path))
        if "Success" not in output:
            self.set_error(output)
            logger.error(output)
            return False
        else:
            return True

    def force_install(self, filename):
        cmd = " install -r -d "+filename
        output = self.run_adb(cmd)
        if "Success" not in output:
            self.set_error(output)
            logger.error(output)
            return False
        else:
            return True

    def uninstall(self, pkg_name):
        cmd = "uninstall "+pkg_name
        output = self.run_adb(cmd)
        return output

    def screen(self, local, display_id=None):
        if display_id is not None:
            raw_data = self.run_shell("screencap -d %d -p" % display_id)
            raw_data = raw_data.replace(self.line_breaker, b"\n")
            f = open(local, "wb")
            f.write(raw_data)
            f.close()
        else:
            filename = self.get_device_timefile()+".png"
            remote = self.screen_cap(filename)
            self.pull(remote, local)
            return filename

    def get_device_timefile(self):
        return self.get_device_time("%Y_%m_%d__%H_%M_%S")

    def get_device_time(self, fmt="%Y%m%d"):
        cmd = ' date "+%s"' % fmt
        return self.run_shell(cmd).strip()

    def get_time_delta(self):
        """
        近似计算运行的电脑跟手机直接的时间差值，电脑时间-手机时间
        :return:
        """
        fmt = "%Y%m%d %H:%M:%S"
        pc_now = datetime.datetime.now()
        ds = self.get_device_time(fmt)
        mobile_now = datetime.datetime.strptime(ds, fmt)
        return pc_now - mobile_now

    def pull_installed_apk(self, pkg, pc_path):
        output = self.run_shell("pm path %s" % pkg)
        path = output.strip().split(":")[-1]
        if not path:
            return False
        else:
            return self.pull(path, pc_path)

    def assert_file_exists(self, filename):
        output = self.run_shell(" ls "+filename )
        if "No such" in output:
            raise AdbException(output)

    def file_exists(self, filename):
        output = self.run_shell(" ls "+filename)
        if "No such" in output or "No such" in self._error_msg:
            return False
        else:
            return True

    def file_size(self, filepath):
        out = self.run_shell("stat -c %s {0}".format(filepath))
        if out.isdigit():
            file_size = int(out)
        else:
            file_size = 0
        return file_size

    def get_newest_filename(self, remote_dir):
        """取指定目录下最新生成的文件"""
        self.assert_file_exists(remote_dir)
        output = self.run_shell('ls ' + remote_dir + ' -t | head -1')
        return output.strip() if output and output.strip() else None

    def apk_install_info(self, pkgname):
        output = self.run_shell("pm path com.tencent.mm")
        m = re.search(r"(/\S+)+", output)
        path = m.group()
        logger.info("apk path:"+path)
        output = self.run_shell("ls -l "+path)
        m = re.search(r"\d+", output)
        return m.group(), self.get_file_modifytime(path)

    @decorator.cached_property
    def is_wetest_custom_device(self):
        model = self.get_property("ro.product.model")
        if model and "wetest" in model.lower():
            # WeTest定制手机
            return True
        return False

    @decorator.cached_property
    def is_emulator(self):
        """
        判断是否是模拟器
        :return: 
        """
        if self.is_wetest_custom_device:
            return False
        qemu = self.get_property("ro.kernel.qemu")
        if qemu and qemu != '0':
            return True
        else:
            return False

    @decorator.cached_property
    def line_breaker(self):
        if self.sdk_version >= 24:
            line_breaker = os.linesep
        else:
            line_breaker = '\r' + os.linesep
        return line_breaker.encode("ascii")

    def check_installed_apk(self, pkg, apk_path, install=False):
        """
        检查安装的apk的版本是否匹配
        :param pkg: 对比的apk的包名
        :param apk_path: 对比的apk路径
        :param install:  如果不匹配，是否重新安装
        :return:
        """
        temp_dir = tempfile.gettempdir()
        temp_apk_path = os.path.join(temp_dir, "temp.apk")
        ret = False
        if self.pull_installed_apk(pkg, temp_apk_path):
            if at.utils.apkinfo.ApkInfo(apk_path) != at.utils.apkinfo.ApkInfo(temp_apk_path):
                if install:
                    ret = self.install(apk_path)
            else:
                ret = True
        else:
            logger.debug("%s not installed", pkg)
            if install:
                ret = self.install(apk_path)
        logger.info("%s check result:%s", pkg, ret)
        if os.path.exists(temp_apk_path):
            os.remove(temp_apk_path)
        return ret

    def push_media(self, src_file_path, dest_file_path):
        '''
        src_path, dest_path 都是完整的文件全路径
        '''
        self.push(src_file_path, dest_file_path)
        self.broadcast("android.intent.action.MEDIA_SCANNER_SCAN_FILE -d file://%s" % dest_file_path)

    def push_medias(self, src_dir_path, dest_dir_path):
        """
        推送图片视频文件夹到手机
        :param src_dir_path: 本地目录
        :param dest_dir_path: 手机上的目录
        :return:
        """
        f_list = []
        for root, dirs, filenames in os.walk(src_dir_path):
            for f in filenames:
                f_list.append(os.path.join(root, f))
        for src_f in f_list:
            f_name = os.path.basename(src_f)
            dest_f = os.path.join(dest_dir_path, f_name)
            self.push_media(src_f, dest_f)


if "__main__" == __name__:
    adb = AdbWrap.apply_adb(None)
    print(adb.get_current_activity2())
    # print adb.install("/Users/mmtest/code/MMWebTest/at/bin/AtServer.apk")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Created by xiazeng on 2018/8/23

import logging
from .utils import decorator
from .core import uixml

logger = logging.getLogger()


class AtHook(object):
    def __init__(self):
        self.frame_changed_callbacks = []
        self._has_ran = False

    def before_operation(self):
        pass

    def register_frame_changed(self, method):
        self.frame_changed_callbacks.append(method)

    @decorator.cost_log
    def frame_changed(self, rate, last_views, current_views, act_name):
        """
        界面变化
        :param rate: 相似性(0~1)
        :param last_views: 上一个界面的元素
        :param current_views: 当前界面的元素
        :param act_name: Activtiy类名
        :type last_views: list[uixml.UiView]
        :type current_views: list[UiXml.UiView]
        :return: 
        """
        if self._has_ran:
            return
        self._has_ran = True
        for cb in self.frame_changed_callbacks:
            cb(rate, last_views, current_views, act_name)
        self._has_ran = False



# -*- coding: utf-8 -*-
# Created by xiazeng on 2019/12/27


class AtError(Exception):
    pass


class AtUnknownError(AtError):
    pass


class FailedConnectAtServer(AtError):
    pass


class AdbException(AtError):
    pass


class AdbDisconnectedError(AdbException):
    pass


class CanNotAccessToPackageManager(AtError):
    pass


class RemoteError(AtError):
    pass


class UiUnExpectError(RemoteError):
    pass


class UiNotFoundError(RemoteError):
    pass


class NoSuchMethodError(RemoteError):
    pass


class ParamError(RemoteError):
    pass


class ClickPermissionError(RemoteError):
    pass


class UiAutomatorDisconnectError(RemoteError):
    pass


'''
Author: yopofeng yopofeng@tencent.com
Date: 2023-03-06 11:38:49
LastEditors: yopofeng yopofeng@tencent.com
LastEditTime: 2023-03-14 17:45:53
FilePath: /py-minium/minium/utils/meta.py
Description: 定义一些监控类
'''


class getter:
    def __init__(self, src) -> None:
        self.__src = src

    def __call__(self, _self):
        return getattr(_self, self.__src)


class setter:
    def __init__(self, src, callback=None) -> None:
        self.__src = src
        self.__callback = callback

    def __call__(self, _self, _value):
        if _value != getattr(_self, self.__src) and self.__callback:
            self.__callback(_self, _value)
        setattr(_self, self.__src, _value)


class PropertyMeta(type):
    """属性监听
    监听类中定义以"_"开头属性, 当属性值前后有变化, 如果存在"on_{prop}_change"方法, 则触发
    """

    def __new__(cls, name, base, attr_dict):
        keys = attr_dict.keys()
        for k in list(keys):
            if k.startswith("__"):
                continue
            if k.startswith("_"):
                prop_name = k[1:]
                callback = attr_dict.get(f"on_{prop_name}_change", None)
                attr_dict[prop_name] = property(
                    getter(k), setter(k, callable(callback) and callback)
                )
        return super().__new__(cls, name, base, attr_dict)


if __name__ == "__main__":
    class A(object, metaclass=PropertyMeta):
        _id = None
        def __init__(self) -> None:
            pass

    class B(A):
        _id = None  # 需要监听的属性必须需要在cls下有定义
        def on_id_change(self, value):
            print("B.id change", value)

    b = B()
    b.id = 123


#!/usr/bin/env python3

"""
Author:         xiazeng
Create time:    2021/5/20
Description:    

"""
import logging
import unittest

logging.basicConfig(level=logging.DEBUG,
                    format="%(levelname)-5.5s %(asctime)s %(filename)s %(funcName)s %(lineno)-3d %(message)s")

logger = logging.getLogger()

from airtest.core.api import *


class AirTestTest(unittest.TestCase):
    def setUp(self) -> None:
        init_device()

    def test_login(self):
        touch((100, 100))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@Author: lockerzhang
@LastEditors: lockerzhang
@Description: 小程序页面
@Date: 2019-03-11 14:42:29
@LastEditTime: 2019-06-05 17:04:02
"""
from __future__ import annotations
import re
from typing_extensions import Self
from minium.framework.exception import MiniElementNotFoundError, MiniAppError, MiniTimeoutCauseByConnectionBreakError
import typing
from .element import *
from .minium_object import MiniumObject, RetryableApi
from ...utils.utils import timeout
from .connection import Command
import json
if typing.TYPE_CHECKING:
    from .app import App

# 需要在setup中加上cssselect库
# from cssselect.xpath import HTMLTranslator
# from cssselect.parser import SelectorError

# translator = HTMLTranslator()


class Page(MiniumObject, metaclass=RetryableApi):
    """
    页面相关接口
    """
    __RETRY_EXCEPTION__ = (MiniTimeoutCauseByConnectionBreakError,)  # 小程序掉线Page实例可能就没用了
    __RETRY_API__ = [
        "data",
        "inner_size",
        "scroll_height",
        "scroll_width",
        "scroll_x",
        "scroll_y",
        "scroll_to",
        "_get_window_properties",
        # "element_is_exists",
        # "get_element",
        # "get_elements",
        # "_element_is_exists",
        # "get_elements_by_xpath",
        # "get_element_by_xpath",
        # element 相关底层方法
        "__get_elements",
        "_get_element",
        "_get_element_by_xpath",
        "_get_elements_by_xpath",
    ]

    PLUGIN_PATH_REG = r"^(plugin://|plugin-private://|/?__plugin__/)(\w+)/"

    def __init__(self, page_id, path, query, app: App):
        """
        初始化页面
        """
        super().__init__()
        self.page_id = page_id
        # 处理下path，普通page以"/"开头，插件以"plugin://" 或 "plugin-private://" 或 "__plugin__" 开头
        self.plugin_appid = None
        match = re.search(Page.PLUGIN_PATH_REG, path)
        if match:
            self.plugin_appid = match.group(2)
        elif not path.startswith("/"):
            path = "/" + path
        self.path = path
        self.query = query
        self.app = app
        if app:
            self.connection = app.connection
        self.is_webview = None  # None: 未知，true: 是，false: 否

    def __del__(self):
        self.app = None

    def __repr__(self):
        return "Page(id={0}, path={1}, query={2})".format(
            self.page_id, self.path, self.query
        )

    def __eq__(self, page: Self or str):
        """
        重载 ==
        直接对比page path和query
        """
        if isinstance(page, Page):
            # 都有page id的情况, 直接判断id即可
            if page.page_id and self.page_id:
                return page.page_id == self.page_id
            # query都不为None, 进行query的对比
            if page.query is not None and self.query is not None:
                if json.dumps(page.query) != json.dumps(self.query):
                    return False
            # 不同页面类型
            if self.plugin_appid != page.plugin_appid:
                return False
            # 对比 path
            if self.plugin_appid:
                # 插件页面需要处理一下path
                self_path = re.sub(Page.PLUGIN_PATH_REG, "", self.path)
                page_path = re.sub(Page.PLUGIN_PATH_REG, "", page.path)
                return self_path == page_path
            return self.path == page.path
        else:
            _page = Page(None, page, None, None)
            return self == _page

    def __ne__(self, page: Self or str) -> bool:
        return not (self == page)

    @property
    def data(self):
        """
        获取页面 Data
        :return: json
        """
        return self._send("Page.getData").result.data

    @data.setter
    def data(self, data):
        """
        设置页面 data
        :param data:
        :return:
        """
        self._send("Page.setData", {"data": data})

    @property
    def inner_size(self):
        """
        get window size
        :return:
        """
        size_arr = self._get_window_properties(["innerWidth", "innerHeight"])
        return {"width": size_arr[0], "height": size_arr[1]}

    @property
    def scroll_height(self):
        """
        get scroll height
        :return:
        """
        return self._get_window_properties(["document.documentElement.scrollHeight"])[0]

    @property
    def scroll_width(self):
        """
        get scroll width
        :return:
        """
        return self._get_window_properties(["document.documentElement.scrollWidth"])[0]

    @property
    def scroll_x(self):
        """
        获取窗口顶点与页面顶点的 x 轴偏移量
        :return:
        """
        return self._get_window_properties(["scrollX"])[0]

    @property
    def scroll_y(self):
        """
        获取窗口顶点与页面顶点的 y 轴偏移量
        :return:
        """
        return self._get_window_properties(["scrollY"])[0]

    def wait_data_contains(self, *keys_list: list or str, max_timeout: int = 10):
        """
        description: 等待Page.data中包含指定keys
        param {*} self
        param {array} keys_list: items is list or str split by "."
        param {int} max_timeout
        return {bool}
        """

        @timeout(max_timeout)
        def f():
            d = self.data
            for keys in keys_list:
                obj = d
                for key in keys if isinstance(keys, (list, tuple)) else keys.split("."):
                    if isinstance(obj, dict) and key in obj:
                        obj = obj[key]
                    else:
                        return False
            return True

        try:
            return f()
        except Exception as e:
            self.logger.exception(e)
            return False

    def element_is_exists(
        self,
        selector: str = None,
        max_timeout: int = 10,
        inner_text=None,
        text_contains=None,
        value=None,
        xpath: str = None,
    ) -> bool:
        """
        查询元素是否存在
        :param selector:
        :param max_timeout: 超时时间
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        :param xpath: 使用xpath
        :return: bool
        """
        if selector and selector.startswith("/"):
            # 以/或//开头的认为是xpath
            xpath = selector
            selector = None
        return self._wait(
            lambda: self._element_is_exists(
                selector,
                xpath,
                inner_text=inner_text,
                value=value,
                text_contains=text_contains,
            ),
            max_timeout,
        )

    def get_element(
        self,
        selector: str,
        inner_text=None,
        text_contains=None,
        value=None,
        max_timeout=0,
        xpath=None,
    ) -> ElementType:
        """
        find elements in current page, by css selector or xpath
        目前支持的css选择器有:

        选择器	              样例	                      样例描述
        .class	             .intro	                    选择所有拥有 class="intro" 的组件
        #id	                 #firstname	                选择拥有 id="firstname" 的组件
        tagname	             view	                    选择所有    view 组件
        tagname, tagname	 view, checkbox	            选择所有文档的 view 组件和所有的 checkbox 组件

        ::after	             view::after	            在 view 组件后边插入内容
        ::before	         view::before	            在 view 组件前边插入内容
        >>>                  ce1>>>.ce2>>>.intro        跨自定义组件的后代选择器

        支持以/或//开头的xpath

        :param selector: CSS选择器/XPATH
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        :param max_timeout: 超时时间
        :return:element 对象
        """
        if (selector and selector.startswith("/")) or xpath:
            # use xpath
            return self.get_element_by_xpath(
                xpath or selector,
                max_timeout=max_timeout,
                inner_text=inner_text,
                value=value,
                text_contains=text_contains,
            )
        r = self.get_elements(
            selector,
            max_timeout,
            inner_text=inner_text,
            value=value,
            text_contains=text_contains,
            index=0,
        )
        if not r:
            raise MiniElementNotFoundError("element[%s] not found" % selector)
        return r[0]

    def call_method(self, method, args=None):
        if not args:
            args = []
        if isinstance(args, dict):
            args = [args]
        return self._send("Page.callMethod", {"method": method, "args": args})

    def call_function(self, app_function: str, args=None, sync=True, desc=None):
        """
        向 webview 页面注入代码并执行, 仅支持web-view页面
        :param app_function:
        :param args:
        :param sync:
        :param desc: 报错描述
        :return:
        """
        if self.is_webview is None:
            # 检查是否为webview页面
            if self.get_elements("web-view"):
                self.is_webview = True
            else:
                self.is_webview = False
        if self.is_webview:
            return self._page_evaluate(app_function=app_function, args=args, sync=sync)
        return self._evaluate(
            app_function=app_function, args=args, sync=sync, desc=desc
        )

    def wait_for(self, condition=None, max_timeout=10):
        s_time = time.time()
        if isinstance(condition, int):
            time.sleep(condition)
            self.logger.debug("waitFor: %s s" % (time.time() - s_time))
            return True
        elif isinstance(condition, str):
            while (time.time() - s_time) < max_timeout:
                if self._element_is_exists(condition):
                    return True
                else:
                    time.sleep(0.25)
            else:
                return False
        elif hasattr(condition, "__call__"):
            while (time.time() - s_time) < max_timeout:
                res = condition()
                if res:
                    return True
                else:
                    time.sleep(0.25)
            else:
                return False

    def get_elements(
        self,
        selector,
        max_timeout=0,
        inner_text=None,
        text_contains=None,
        value=None,
        index=-1,
        xpath=None,
    ) -> typing.List[ElementType]:
        """
        find elements in current page, by css selector
        :param selector: 选择器
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        :param max_timeout: 超时时间
        :param index: index == -1: get所有, index >=0: get index+1个
        :param xpath: 使用xpath
        :return:element 对象 list
        """
        if selector and selector.startswith("/"):
            # 以/或//开头的认为是xpath
            xpath = selector
            selector = None
        if inner_text is None and value is None and text_contains is None:
            need_filter = False
        else:
            need_filter = True
        if not need_filter and selector:
            # 不需要过滤内容，直接返回
            return self._get_elements_by_css(selector, max_timeout, index=index)
        elif xpath:
            # xpath支持text() contains()等条件，不需要额外做过滤
            return self.get_elements_by_xpath(
                xpath,
                max_timeout=max_timeout,
                inner_text=inner_text,
                text_contains=text_contains,
                value=value,
            )

        @timeout(max_timeout)
        def filter_elements():
            # 需要过滤内容，有返回元素不是终结条件, 需要获取所有元素后再作过滤
            elements = self._get_elements_by_css(selector, max_timeout, index=-1)
            els = []
            for element in elements:
                if inner_text and element.inner_text != inner_text:
                    continue
                if value and element.value() != value:
                    continue
                if text_contains and text_contains not in element.inner_text:
                    continue
                els.append(element)
                if len(els) == (index + 1):
                    return els
            return els

        return filter_elements()

    def scroll_to(self, scroll_top, duration=300):
        """
        滚动到指定位置
        :param scroll_top:  位置 px
        :param duration:  滚动时长
        :return:
        """
        if not self.app:
            raise RuntimeError("please instantiate Page through App.get_current_page()")
        self.app.call_wx_method(
            "pageScrollTo", [{"scrollTop": scroll_top, "duration": duration}]
        )

    def get_elements_by_xpath(
        self,
        xpath,
        max_timeout=10,
        inner_text=None,
        value=None,
        text_contains=None,
        index=-1,
    ) -> ElementType:
        """
        根据xpath查找元素
        :param xpath: xpath
        :param max_timeout: 超时时间
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        """
        if inner_text is not None:
            xpath += '[text()="%s"]' % inner_text
        elif text_contains is not None:
            xpath += '[contains(text(), "%s")]' % text_contains
        elif value is not None:
            xpath += '[@value="%s"]' % value

        @timeout(max_timeout)
        def search_elements():
            return self._get_elements_by_xpath(xpath, index=index)

        return search_elements()

    def get_element_by_xpath(
        self, xpath, max_timeout=10, inner_text=None, value=None, text_contains=None
    ) -> ElementType:
        """
        根据xpath查找元素
        :param xpath: xpath
        :param max_timeout: 超时时间
        :param inner_text: inner_text
        :param value: value
        :param text_contains: 包含的文字
        """
        if inner_text is not None:
            xpath += '[normalize-space(text())="%s"]' % inner_text
        elif text_contains is not None:
            xpath += '[contains(text(), "%s")]' % text_contains
        elif value is not None:
            xpath += '[@value="%s"]' % value

        @timeout(max_timeout)
        def search_element():
            return self._get_element_by_xpath(xpath)

        el = search_element()
        if not el:
            raise MiniElementNotFoundError("element[%s] not found" % xpath)
        return el

    def _get_window_properties(self, names=None):
        """
        获取 window 对象的属性值。
        :param names:
        :return:
        """
        if names is None:
            names = []
        return self._send(
            "Page.getWindowProperties", {"names": names}
        ).result.properties

    def _send(self, method, params=None, sync=True):
        if params is None:
            params = {}
        params["pageId"] = self.page_id
        # self.logger.debug(f"method {method}, params: {params}")
        return (
            self.connection.send(method, params)
            if sync
            else self.connection.send_async(method, params)
        )

    def __get_elements(self, selector: str, index=-1):
        elements = []
        ret = self._send("Page.getElements", {"selector": selector})
        for el in ret.result.elements:
            elements.append(create(el, self.page_id, self.connection))
            if len(elements) == (index + 1):  # index==-1时，不会match，就会全部返回
                return elements
        return elements

    def __search_child(
        self, selector_list: list, parent: ElementType = None, index=-1
    ) -> typing.List[ElementType]:
        # index == -1: get所有, index >=0: get index+1个
        if len(selector_list) == 0:
            return []
        _selector = selector_list.pop()
        should_be_custom_element = bool(
            len(selector_list)
        )  # 出了最后一层selector，都需要是自定义组件，不然不能往下走
        if parent:
            els = parent.get_elements(
                _selector, max_timeout=0
            )  # 调用element.get_elements方法，要求立刻返回，不然会影响效率
        else:
            els = self.__get_elements(_selector)
        if len(els) == 0:
            return []
        real_els = []
        for _el in els:
            if should_be_custom_element and isinstance(_el, CustomElement):
                real_els.append(_el)
            elif not should_be_custom_element:
                real_els.append(_el)
            else:
                self.logger.warn("%s should be a custom element" % _selector)
        if not real_els or len(selector_list) == 0:  # 找不到 或 没有子选择器了
            return real_els if index <= -1 else real_els[:index+1]
        child_els = []
        for _el in real_els:
            child_el = self.__search_child(
                selector_list[0:], _el, index=-1 if index <= -1 else index - len(child_els)
            )  # selector_list[0:]相当于copy
            child_els += child_el
            if len(child_els) == (index + 1):
                return child_els
        return child_els

    def _get_elements_by_css(
        self, selector: str, max_timeout=0, index=-1
    ) -> typing.List[ElementType]:
        """
        1. 现存自定义组件中的class会自动加前缀，但不包括slot占位的元素
        2. slot元素的class命名规则
        2.1 <page><test><view class="这个class不会加前缀"></view></test></page>
        2.2 <custom><test><view class="这个class会加上custom组件对应的前缀"></view></test></custom>
        3. 自定义组件中通过id获取元素失败
        """

        @timeout(max_timeout)
        def search_elements(_selector: list or tuple):
            return self.__search_child(
                _selector[0:], index=index
            )  # __search_child回pop元素，第一次search失败后重试会有问题，copy一份

        self.logger.info("try to get elements: %s" % selector)
        _selector_list = selector.split(">>>")
        _selector_list.reverse()  # 新增处理【>>>】穿透自定义组件逻辑
        els = search_elements(_selector_list)
        if len(els) == 0:
            self.logger.warning(f"Could not found any element '{selector}' you need")
        else:
            self.logger.info("find elements success: %s" % str(els))
        return els

    def _get_element(self, selector, max_timeout=0) -> ElementType:
        def search_element():
            ret = self._send("Page.getElement", {"selector": selector})
            return create(ret.result, self.page_id, self.connection)

        self.logger.info("try to get element: %s" % selector)
        el = search_element()
        self.logger.info("find element success: %s" % str(el))
        return el

    def _element_is_exists(
        self, selector: str = None, xpath: str = None, **kwargs
    ) -> bool:
        """
        description:
        param {*} self
        param {str} selector: css 选择器
        param {str} xpath: xpath
        param {object} kwargs: inner_text, text_contains, value ...
        return {bool}
        """
        if selector and xpath:
            self.logger.warning("selector and xpath both not None, use selector")
        if not (selector or xpath):
            raise RuntimeError("Must use either selector or xpath")
        try:
            if selector:
                return (
                    True if self.get_elements(selector, 0, index=0, **kwargs) else False
                )
            else:
                return (
                    True
                    if self.get_elements_by_xpath(xpath, 0, index=0, **kwargs)
                    else False
                )
        except Exception:
            return False

    # 正式支持xpath后，考虑复用get_element接口，通过检测selector类型来决定使用什么选择器
    # def _is_xpath(self, selector):
    #     """
    #     检测一个selector是否是xpath
    #     1. start with "/"
    #     2. start with "//"
    #     3. start with "./" or == "."
    #     4. translator.css_to_xpath(css_selector) fail
    #     """
    #     if selector.startswith("/") or selector.startswith("//") or selector.startswith("./"):
    #         return True
    #     if selector == ".":
    #         return True
    #     try:
    #         translator.css_to_xpath(selector)
    #         return False
    #     except SelectorError:
    #         return True
    #     return False

    def _get_element_by_xpath(self, xpath: str) -> ElementType:
        """
        description: 通过xpath获取元素
        param {*} self
        param {str} xpath
        return {*} ElementType or None
        """
        xpath = (
            xpath[5:] if xpath.find("/page/") == 0 else xpath
        )  # xpath不能以/page/开头（不存在的根节点）
        try:
            ret = self._send("Page.getElementByXpath", {"selector": xpath})
        except MiniAppError as e:
            if str(e) == "no such element":
                return None
            raise
        return create(ret.result, self.page_id, self.connection)

    def _get_elements_by_xpath(self, xpath: str, index=-1) -> typing.List[ElementType]:
        """
        description: 通过xpath获取元素
        param {*} self
        param {str} xpath
        return {*} List[ElementType]
        """
        xpath = (
            xpath[5:] if xpath.find("/page/") == 0 else xpath
        )  # xpath不能以/page/开头（不存在的根节点）
        if not self._can_i_use("Page.getElementsByXpath"):
            el = self._get_element_by_xpath(xpath)  # 降级使用Page.getElementByXpath
            return [el] if el else []
        try:
            ret = self._send("Page.getElementsByXpath", {"selector": xpath})
        except MiniAppError as e:
            if str(e) == "no such element":
                return []
            if (
                str(e).find("Page.getElementsByXpath unimplemented") > 0
            ):  # or str(e) in ["r is not iterable",]:
                self._unset_interface("Page.getElementsByXpath")
                el = self._get_element_by_xpath(xpath)  # 降级使用Page.getElementByXpath
                return [el] if el else []
            raise
        elements = []
        for el in ret.result.elements:
            elements.append(create(el, self.page_id, self.connection))
            if len(elements) == (index + 1):
                return elements
        return elements

    def _page_evaluate(self, app_function: str, args=None, sync=True):
        """
        在当前页面执行命令
        """
        if not args:
            args = []
        return self._send(
            "Page.callFunction",
            {"functionDeclaration": app_function, "args": args},
            sync=sync,
        )
